/*
linux/arch/m68k/atari/atasound.c

VoxWare compatible Atari TT DMA sound driver for 680x0 Linux

(c) 1995 by Michael Schlueter & Michael Marte

Michael Schlueter (michael@duck.syd.de) did the basic structure of the VFS
interface and the u-law to signed byte conversion.

Michael Marte (marte@informatik.uni-muenchen.de) did the sound queue,
/dev/mixer, /dev/sndstat and complemented the VFS interface. He would like
to thank:
Michael Schlueter for initial ideas and documentation on the MFP and
the DMA sound hardware.
Therapy? for their CD 'Troublegum' which really made me rock.

The author of atari_nosound, atari_mksound and atari_microwire_cmd is
unknown.
(++roman: That's me... :-)

/dev/sndstat is based on code by Hannu Savolainen, the author of the
VoxWare family of drivers.

This file is subject to the terms and conditions of the GNU General Public
License.  See the file README.legal in the main directory of this archive
for more details.

History:
1995/8/25	first release

1995/9/02	++roman: fixed atari_stram_alloc() call, the timer programming and
			several race conditions

1995/9/14	++roman: After some discussion with Michael Schlueter, revised the
			interrupt disabling
			Slightly speeded up U8->S8 translation by using long operations
			where possible
			Added 4:3 interpolation for /dev/audio

1995/9/20	++TeSche: fixed a bug in sq_write and changed /dev/audio
			converting to play at 12517Hz instead of 6258Hz

1995/9/23	++TeSche: Changed sq_interrupt() and sq_play() to pre-program
			the DMA for another frame while there's still one
			running. This allows the IRQ response to be arbitrarily
			delayed and playing will still continue.
*/


#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/mm.h>

#include <asm/atarihw.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/atariints.h>
#include <asm/bootinfo.h>

#include "atasound.h" 
#include <linux/soundcard.h>


/*
stuff from the old atasound.c
*/


static void atari_nosound( unsigned long ignored )
{
	unsigned char	tmp;
	
	/* turn off generator A in mixer control */
	sound_ym.rd_data_reg_sel = 7;
	tmp = sound_ym.rd_data_reg_sel;
	sound_ym.wd_data = tmp | 0x39;
}	
		

void atari_microwire_cmd( int cmd )
{
	tt_microwire.mask = 0x7ff;
	tt_microwire.data = MW_LM1992_ADDR | cmd;

	/* Busy wait for data being completely sent :-( */
	while( tt_microwire.mask != 0x7ff)
		;
}


#define	PC_FREQ		1192180
#define	PSG_FREQ	125000


void atari_mksound( unsigned int count, unsigned int ticks )
/*
 * Generates sound of some count for some number of clock ticks
 * [count = 1193180 / frequency]
 */
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						     atari_nosound };
	unsigned long flags;
	unsigned char tmp;

	save_flags(flags);
	cli();

	if (count == 0x637 && ticks == HZ/8) {
		/* Special case: These values are used by console.c to
		 * generate the console bell. They are catched here and the
		 * sound actually generated is somehow special: it uses the
		 * generator B and an envelope. No timer is needed therefore
		 * and the bell doesn't disturb an other ongoing sound.
		 */

		/* set envelope duration to 492 ms */
		sound_ym.rd_data_reg_sel = 11;
		sound_ym.wd_data = 0;
		sound_ym.rd_data_reg_sel = 12;
		sound_ym.wd_data = 15;
		/* envelope form: max -> min single */
		sound_ym.rd_data_reg_sel = 13;
		sound_ym.wd_data = 9;
		/* set generator B frequency to 2400 Hz */
		sound_ym.rd_data_reg_sel = 2;
		sound_ym.wd_data = 52;
		sound_ym.rd_data_reg_sel = 3;
		sound_ym.wd_data = 0;
		/* set volume of generator B to envelope control */
		sound_ym.rd_data_reg_sel = 9;
		sound_ym.wd_data = 0x10;
		/* enable generator B in the mixer control */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x02) | 0x38;

		restore_flags(flags);
		return;
	}

	del_timer( &sound_timer );

	if (!count) {
		atari_nosound( 0 );
	}
	else {

		/* convert from PC counter value (base frequency 1.193
		   MHz) to PSG period value (base frequency 125 kHz).  */
		int period = (PSG_FREQ * count + PC_FREQ/2) / PC_FREQ;

		if (period > 0xfff) period = 0xfff;

		/* set generator A frequency to 0 */
		sound_ym.rd_data_reg_sel = 0;
		sound_ym.wd_data = period & 0xff;
		sound_ym.rd_data_reg_sel = 1;
		sound_ym.wd_data = (period >> 8) & 0xf;
		/* turn on generator A in mixer control (but not noise
		 * generator!) */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x01) | 0x38;
		/* set generator A level to maximum, no envelope */
		sound_ym.rd_data_reg_sel = 8;
		sound_ym.wd_data = 15;
		
		if (ticks) {
			sound_timer.expires = ticks;
			add_timer( &sound_timer );
		}
	}

	restore_flags(flags);
}


#ifdef CONFIG_ATARI_DMA_SOUND

static int ioctl_return (int *addr, int value);
#define IOCTL_IN(arg)		get_fs_long ((long *)(arg))
#define IOCTL_OUT(arg, ret)	ioctl_return ((int *)arg, ret)


/*
low level sound stuff
*/


struct sound_settings {
	int format;	/* AFMT_{MU_LAW, U8, S8} */
	int speed;	/* DMASND_MODE_{6KHZ, 12KHZ, 25KHZ, 50KHZ} */
	int volume_left, volume_right;	/* [-40, 0] db */
	int bass, treble;	/* [-12, 12] db */
	int stereo;	/* DNASND_MODE_{MONO, STEREO} */
};


struct sound_settings sound;


static void sound_silence (void)
{
	atari_microwire_cmd (MW_LM1992_PSG_HIGH);
		/* mix in PSG signal 1:1 */
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
}


static void sound_init (void)
{
	sound_silence ();
	sound.stereo = DMASND_MODE_MONO;
	sound.speed = DMASND_MODE_6KHZ;
	tt_dmasnd.mode =  DMASND_MODE_MONO | DMASND_MODE_6KHZ;
}


static int sound_set_speed (int speed)
{
	if (speed > 40000) {
		speed = 50066;
		sound.speed = DMASND_MODE_50KHZ;
	} else if (speed > 20000) {
		speed = 25033;
		sound.speed = DMASND_MODE_25KHZ;
	} else if (speed > 7000)  {
		speed = 12517;
		sound.speed = DMASND_MODE_12KHZ;
	} else {
		speed = 6258;
		sound.speed = DMASND_MODE_6KHZ;
	}

	tt_dmasnd.mode = sound.stereo | sound.speed;

	return speed;
}


static int sound_set_format (int format)
{
	if (format != AFMT_MU_LAW && format != AFMT_U8 && format != AFMT_S8)
	  format = AFMT_U8;
	sound.format = format;
	return format;
}



static int sound_set_stereo (int stereo)
{
	if (stereo != 0) stereo = 1;
	sound.stereo = (stereo == 1)? DMASND_MODE_STEREO: DMASND_MODE_MONO;
	tt_dmasnd.mode = sound.stereo | sound.speed;
	return stereo;	
}



#define VOLUME_VOXWARE_TO_DB(v) (((v) < 0)? -40: ((v) > 100)? 0: (((v) * 41 + 50) / 100) - 40)
#define VOLUME_DB_TO_VOXWARE(v) ((((v) + 40) * 100 + 20) / 41)
#define TONE_VOXWARE_TO_DB(v) (((v) < 0)? -12: ((v) > 100)? 12: (((v) + 2) / 4) - 12)
#define TONE_DB_TO_VOXWARE(v) (((v) + 12) * 4)


static int sound_set_volume (int volume)
{
	sound.volume_left = VOLUME_VOXWARE_TO_DB (volume & 0xff);
	atari_microwire_cmd (MW_LM1992_BALLEFT (sound.volume_left));
	sound.volume_right = VOLUME_VOXWARE_TO_DB ((volume & 0xff00) >> 8);
	atari_microwire_cmd (MW_LM1992_BALRIGHT (sound.volume_right));
	return VOLUME_DB_TO_VOXWARE (sound.volume_left) |
		(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8);
}


static int sound_set_treble (int treble)
{
	sound.treble = TONE_VOXWARE_TO_DB (treble);
	atari_microwire_cmd (MW_LM1992_TREBLE (sound.treble));
	return TONE_DB_TO_VOXWARE (sound.treble);
}


static int sound_set_bass (int bass)
{
	sound.bass = TONE_VOXWARE_TO_DB (bass);
	atari_microwire_cmd (MW_LM1992_BASS (sound.bass));
	return TONE_DB_TO_VOXWARE (sound.bass);
}


static unsigned char ulaw_2_dma [] = {
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -126, 
  -123, -119, -115, -111, -107, -103,  -99,  -95, 
   -91,  -87,  -83,  -79,  -75,  -71,  -67,  -63, 
   -60,  -58,  -56,  -54,  -52,  -50,  -48,  -46, 
   -44,  -42,  -40,  -38,  -36,  -34,  -32,  -30, 
   -28,  -27,  -26,  -25,  -24,  -23,  -22,  -21, 
   -20,  -19,  -18,  -17,  -16,  -15,  -14,  -13, 
   -13,  -12,  -12,  -11,  -11,  -10,  -10,   -9, 
    -9,   -8,   -8,   -7,   -7,   -6,   -6,   -5, 
    -5,   -5,   -4,   -4,   -4,   -4,   -3,   -3, 
    -3,   -3,   -2,   -2,   -2,   -2,   -1,   -1, 
    -1,   -1,   -1,   -1,    0,    0,    0,    0, 
     0,    0,    0,    0,    0,    0,    0,    0, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   124,  120,  116,  112,  108,  104,  100,   96, 
    92,   88,   84,   80,   76,   72,   68,   64, 
    61,   59,   57,   55,   53,   51,   49,   47, 
    45,   43,   41,   39,   37,   35,   33,   31, 
    29,   28,   27,   26,   25,   24,   23,   22, 
    21,   20,   19,   18,   17,   16,   15,   14, 
    14,   13,   13,   12,   12,   11,   11,   10, 
    10,    9,    9,    8,    8,    7,    7,    6, 
     6,    6,    5,    5,    5,    5,    4,    4, 
     4,    4,    3,    3,    3,    3,    2,    2, 
     2,    2,    2,    2,    1,    1,    1,    1, 
     1,    1,    1,    1,    0,    0,    0,    0, 
};


static void
sound_copy_translate (unsigned char *buff, const unsigned char *src,
		      unsigned long n)
{
	long tmp1;
	int bal = 0;
	unsigned char c = 0;	/* to keep gcc happy */

	switch (sound.format) {

	case AFMT_MU_LAW:
		/* mu-law is for Sun-compatible /dev/audio whose samples always
		 * have a frequency of 8 kHz, but we just can only play either
		 * 6258Hz or 12517Hz. :(
		 *
		 * ++TeSche: From the two possible solutions we choose to
		 * expand data to the higher sample rate. Expanding is done
		 * by simply doubling every second sample to keep this loop
		 * short and quick. This yields a sample rate of 12000Hz
		 * so samples will be played 4.3% too fast. I heavily
		 * doubt that any further mathematics like using an exact
		 * ratio or not simply doubling sampling but really inter-
		 * polating them will be worth the effort. (This loop can
		 * surely be optimized in assembler too, especially the
		 * get_user_byte() construct translates silly...)
		 */
		while(n--) {
			if ((bal -= 2) < 0) {
				bal += 3;
				c = ulaw_2_dma[get_user_byte(src++)];
			}
			*buff++ = c;
		}
		break;

	case AFMT_U8:
		memcpy_fromfs( buff, src, n );
		/* To convert from unsigned to signed data, we would have to subtract
		 * 128 from each byte. But this is easier done by toggling bit 7,
		 * since we do not care for overflows. And eor can be done on 32 bit
		 * units, which is much faster.
		 */
		if (n >= 4) {
#if 0
			if ((unsigned long)buff & 1)
				*buff++ ^= 0x80, --n;
			if ((unsigned long)buff & 2)
				*((unsigned short *)buff)++ ^= 0x8080, n -= 2;
			m = n/4;
			while( m-- )
				*((unsigned long *)buff)++ ^= 0x80808080;
#else
			/* The code for the above wasn't as bad as the one for the
			 * interpolation, but improvable, too :-) */
			__asm__ __volatile__
				("movel %0,%2		\n\t"
				 "btst  #0,%2		\n\t"
				 "beq   1f			\n\t"
				 "eorb  %3,%0@+		\n\t"
				 "subql	#1,%1		\n"
				 "1:				\n\t"
				 "movel %0,%2		\n\t"
				 "btst  #1,%2		\n\t"
				 "beq   1f			\n\t"
				 "eorw  %3,%0@+		\n\t"
				 "subql	#2,%1		\n"
				 "1:				\n\t"
				 "movel %1,%2		\n\t"
				 "lsrl  #2,%2		\n\t"
				 "beq   1f			\n\t"
				 "subqw #1,%2		\n"
				 "2:				\n\t"
				 "eorl  %3,%0@+		\n\t"
				 "dbra  %2,2b		\n"
				 "1:"
				 : "=&a" (buff), "=&d" (n), "=&d" (tmp1)
				 : "d" (0x80808080), "0" (buff), "1" (n) );
#endif
			n = n%4;
		}
		while( n-- )
			*buff++ ^= 0x80;
		break;

	case AFMT_S8:
		memcpy_fromfs( buff, src, n );
		/* no translation needed */
		break;
		
	}
}

static unsigned long usize2b( unsigned long n ) __attribute__ ((const));
static unsigned long bsize2u( unsigned long n ) __attribute__ ((const));

/* In case of interpolation, return the number of buffer bytes for an amount
 * of user bytes. */

static unsigned long usize2b( unsigned long n )
{
	switch (sound.format) {

	  case AFMT_MU_LAW:
		return n*3/2;
		
	  case AFMT_S8:
	  case AFMT_U8:
	  default:
		return( n );
	}
}

/* The inverse function of the above */

static unsigned long bsize2u( unsigned long n )

{
	switch (sound.format) {

	  case AFMT_MU_LAW:
		return (n*2+2)/3;
		
	  case AFMT_S8:
	  case AFMT_U8:
	  default:
		return( n );
	}
}

/*
/dev/mixer abstraction
*/


struct sound_mixer {
	int busy;
};


struct sound_mixer mixer;


static void mixer_init (void)
{
	mixer.busy = 0;
	atari_microwire_cmd (MW_LM1992_VOLUME (0));
	sound.volume_left = 0;
	atari_microwire_cmd (MW_LM1992_BALLEFT (0));
	sound.volume_right = 0;
	atari_microwire_cmd (MW_LM1992_BALRIGHT (0));
	sound.treble = 0;
	atari_microwire_cmd (MW_LM1992_TREBLE (0));
	sound.bass = 0;
	atari_microwire_cmd (MW_LM1992_BASS (0));
}


static int mixer_open (int open_mode)
{
	if (mixer.busy) return -EBUSY;
	mixer.busy = 1;
	return 0;
}


static int mixer_release (void)
{
	mixer.busy = 0;
	return 0;
}


static int mixer_ioctl (struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch (cmd) {

	case SOUND_MIXER_READ_DEVMASK:
		return IOCTL_OUT (arg, (1 << SOUND_MIXER_VOLUME) | (1 << SOUND_MIXER_TREBLE) | (1 << SOUND_MIXER_BASS));
	case SOUND_MIXER_READ_RECMASK:
		return IOCTL_OUT (arg, 0);
	case SOUND_MIXER_READ_STEREODEVS:
		return IOCTL_OUT (arg, 1 << SOUND_MIXER_VOLUME);

	case SOUND_MIXER_READ_VOLUME:
		return IOCTL_OUT (arg,
			VOLUME_DB_TO_VOXWARE (sound.volume_left) |
			(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8));
	case SOUND_MIXER_READ_BASS:
		return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.bass));
	case SOUND_MIXER_READ_TREBLE:
		return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.treble));

	case SOUND_MIXER_WRITE_VOLUME:
		return IOCTL_OUT (arg, sound_set_volume (IOCTL_IN (arg)));
	case SOUND_MIXER_WRITE_BASS:
		return IOCTL_OUT (arg, sound_set_bass (IOCTL_IN (arg)));
	case SOUND_MIXER_WRITE_TREBLE:
		return IOCTL_OUT (arg, sound_set_treble (IOCTL_IN (arg)));
	
	}

	return -EINVAL;
}




/*
sound queue stuff, the heart of the driver
*/


struct sound_buffer {
	char *start;
	int size;
};


struct sound_queue {
	struct sound_buffer buf;
	int block_size;
	int front, rear, count, max_count;
	int rear_size;
	int playing;
	struct wait_queue *write_queue, *open_queue, *sync_queue;
	int open_mode;
	int busy, syncing;
};


struct sound_queue sq;


#define sq_block_address(i) ((char *)sq.buf.start + (i) * sq.block_size)
#define SIGNAL_RECEIVED (current->signal & ~current->blocked)
#define NON_BLOCKING(open_mode) (open_mode & O_NONBLOCK)
#define ONE_SECOND HZ	/* in jiffies (100ths of a second) */
#define NO_TIME_LIMIT 0xffffffff
#define SLEEP(queue, time_limit) \
	current->timeout = jiffies + (time_limit); \
	interruptible_sleep_on (&queue);
#define WAKE_UP(queue) (wake_up_interruptible (&queue))
#define QUEUE_BLOCK_SIZE 4096
#define SOUND_BUFFER_SIZE (25 * QUEUE_BLOCK_SIZE)

#define	DISABLE_IRQ()			\
	do {				\
		mfp.int_mk_a &= ~0x20;	\
	} while( 0 )

#define	ENABLE_IRQ()			\
	do {				\
		mfp.int_mk_a |= 0x20;	\
	} while( 0 )


static void sq_init (const struct sound_buffer *buf, int block_size)
{
	sq.buf.start = buf->start;
	sq.buf.size = buf->size;
	sq.block_size = block_size;
	sq.front = sq.count = 0;
	sq.rear = -1;
	sq.max_count = buf->size / block_size;
	sq.write_queue = sq.open_queue = sq.sync_queue = 0;
	sq.busy = 0;
	sq.syncing = 0;
	sound_init (); sq.playing = 0;
}


static void sq_play_next_frame (int index)
{
	char *start, *end;

	/* used by sq_play() if all doubts whether there really is
	   something to be played are already wiped out. */
	start = sq_block_address (sq.front);
	end = start + ((sq.count == index)? sq.rear_size: sq.block_size);
	DMASNDSetEnd (VTOP (end));
	DMASNDSetBase (VTOP (start));
	/* Since only an even number of samples per frame can be
	   played, we might loose one byte here. (TO DO) */
	sq.front = (sq.front + 1) % sq.max_count;
	if (++sq.playing == 1)
		tt_dmasnd.ctrl = DMASND_CTRL_ON;
	else /* sq.playing == 2 */
		tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_REPEAT;
}


static void sq_play (void)
{
	/* ++TeSche: Note that sq.playing is no longer just a flag but holds
	   the number of frames the DMA is currently programmed for instead,
	   may be 0, 1 (currently being played) or 2 (pre-programmed).

	   Chances done to sq.count and sq.playing are a bit more subtle again
	   so now I must admit I also prefer disabling the irq here rather than
	   considering all possible situations.  But the point is that
	   disabling the irq doesn't have any bad influence on this version of
	   the driver as we benefit from having pre-programmed the DMA
	   whereever possible: There's no need to reload the DMA at the exact
	   time of an interrupt but only at some time while the pre-programmed
	   frame in playing! */
	DISABLE_IRQ();

	if (sq.playing == 2 ||	/* DMA is 'full' */
	    sq.count <= 0) {	/* nothing to do */
		ENABLE_IRQ();
		return;
	}

	if (sq.playing == 0) {
		/* looks like there's nothing 'in' the DMA yet, so try to put
		   two frames into it (at least one is available). */
		if (sq.count == 1 && sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, the only existing frame is not yet filled and
			   we're not syncing? */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(1);
		if (sq.count == 1) {
			/* no more frames */
			ENABLE_IRQ();
			return;
		}
		if (sq.count == 2 && sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, there were two frames, but the second one is
			   not yet filled and we're not syncing? */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(2);
	} else {
		/* there's already a frame being played so we may only stuff
		   one new into the DMA, but even if this may be the last frame
		   existing the previous one is still on sq.count. */
		if (sq.count == 2 && sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, the only existing frame is not yet filled and
			   we're not syncing? */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(2);
	}
	ENABLE_IRQ();
}


static void sq_interrupt (void)
{
#if 0
	/* if you should want to test this... */
	static int cnt = 0;
	if (sq.playing == 2)
		if (++cnt == 10) {
			/* simulate loosing an interrupt */
			cnt = 0;
			return;
		}
#endif

	/* disable repeats */
	tt_dmasnd.ctrl &= ~DMASND_CTRL_REPEAT;

	if (!sq.playing) {
		/* playing was interrupted and sq_reset() has already cleared
		 * the sq variables, so better don't do anything here.
		 */
		WAKE_UP (sq.sync_queue);
		return;
	}

	/* Probably ;) one frame is finished. Well, in fact it may be that a
	 * pre-programmed one is also finished because there has been a long
	 * delay in interrupt delivery and we've completely lost one, but
	 * there's no way to detect such a situation. In such a case the last
	 * frame will be played more than once and the situation will recover
	 * as soon as the irq gets through.
	 */
	sq.count--;
	sq.playing--;

	WAKE_UP (sq.write_queue);
	/* At least one block of the queue is free now so wake up a writing
	   process blocked because of a full queue.  */

	if ((sq.playing != 1) || (sq.count != 1))
		/* We must be a bit carefully here: sq.count indicates the
		   number of buffers used and not the number of frames to be
		   played.  If sq.count==1 and sq.playing==1 that means the
		   only remaining frame was already programmed earlier (and is
		   currently running) so we mustn't call sq_play() here,
		   otherwise we'll play one frame too much. */
		sq_play();

	if (!sq.playing) WAKE_UP (sq.sync_queue);
	/* We are not playing after sq_play (), so there is nothing to play any
	   more.  Wake up a process waiting for audio output to drain.  */
}


/* ++roman: Note: Since we can interpolate incoming data, we have to
 * distinguish carefully between user byte counts and buffer byte counts. To
 * make this easier, I prefixed the former with "u_", the latter with "b_".
 */

static int sq_write (const char *src, int u_count)
{
	int u_written = 0, b_left, u_left;
	char *dest;

	/* ++TeSche: Is something like this necessary? */
	if (u_count < 1)
		return 0;
	
	/* The interrupt doesn't start to play the last, incomplete frame.
	   Thus, we can append to it without disabling the interrupts!  (Note
	   also, that sq.rear isn't affected by the interrupt.)  */
	if ((sq.count > 0) && ((b_left = sq.block_size - sq.rear_size) > 0)) {
		u_left = bsize2u( b_left );
		if (u_left > u_count) {
			u_left = u_count;
			b_left = usize2b( u_count );
		}
		dest = sq_block_address (sq.rear) + sq.rear_size;
		sound_copy_translate (dest, src, b_left);
		sq.rear_size += b_left;
		u_count -= u_left;
		src += u_left;
		u_written += u_left;
	}

	while (u_count > 0) {
		if (sq.count == sq.max_count) {
			sq_play();
			if (NON_BLOCKING (sq.open_mode))
				return (u_written > 0) ? u_written: -EAGAIN;
			SLEEP (sq.write_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)
				return (u_written > 0) ? u_written: -EINTR;
		} else {
			/* Here, we can avoid disabling the interrupt by first
			   copying and translating the data, and then updating
			   the sq variables.  Until this is done, the interrupt
			   won't see the new frame and we can work on it
			   undisturbed.  */
			b_left = sq.block_size;
			u_left = bsize2u( b_left );
			if (u_left > u_count) {
				u_left = u_count;
				b_left = usize2b( u_count );
			}
			sq.rear = (sq.rear + 1) % sq.max_count;
			dest = sq_block_address (sq.rear);
			sound_copy_translate (dest, src, b_left);
			sq.rear_size = b_left;
			sq.count ++;
			u_count -= u_left;
			src += u_left;
			u_written += u_left;
		}
	}

	sq_play();

	return u_written;
}


static int sq_open (int open_mode)
{
	if (sq.busy) {
		if (NON_BLOCKING (open_mode)) return -EBUSY;
		while (sq.busy) {
			SLEEP (sq.open_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)  return -EINTR;
		}
	}
	sq.open_mode = open_mode;
	sq.busy = 1;
	return 0;
}


static void sq_reset (void)
{
	sound_silence ();
	sq.playing = 0;
	sq.count = 0;
	sq.front = (sq.rear + 1) % sq.max_count;
}


static int sq_sync (void)
{
	int rc = 0;

	sq.syncing = 1;
	sq_play();	/* there may be an incomplete frame waiting */

	while (sq.playing) {
		SLEEP (sq.sync_queue, ONE_SECOND);
		if (SIGNAL_RECEIVED) {
			/* While waiting for audio output to drain, an
			   interrupt occured.  Stop audio output immediately
			   and clear the queue.  */
			sq_reset ();
			rc = -EINTR;
			break;
		}
	}

	sq.syncing = 0;
	return rc;
}


static int sq_release (void)
{
	int rc = 0;
	if (sq.busy) {
		rc = sq_sync ();
		sq.busy = 0;
		WAKE_UP (sq.open_queue);
		/* Wake up a process waiting for the queue beeing released.
		   Note: There may be several processes waiting for a call to
		   open () returning.  */
	}
	return rc;
}


/*
/dev/sndstat
*/


struct sound_state {
	int busy;
	char buf [512];
	int len, ptr;
};


struct sound_state state;


static int state_puts (const char *s)
{
	int l;
	for (l = 0; l < 256 && s [l]; l++);
		/* l = max (256, strlen (s)) */
	if (state.len + l > sizeof (state.buf)) return 0;
	memcpy (&state.buf [state.len], s, l);
	state.len += l;
	return 1;
}


static int state_puti (int i, int radix)
{
	int l;
	static char digit [] = "0123456789abcdef";
	char buf [11];

	l = 0; buf [sizeof (buf) - 1] = 0;

	if (i < 0) {
		state_puts ("-");
		i = -i;
	}

	do {
		buf [sizeof (buf) - 2 - l] = digit [i % radix]; l++;
		i /= radix;
	} while (i);

	return state_puts (&buf [sizeof (buf) - 1 - l]);
}


static void state_init (void)
{
	state.busy = 0;
}


static int state_open (int open_mode)
{
	int	count, rear_size, playing, syncing;
	
	if (state.busy) return -EBUSY;

	state.len = 0;
	state.ptr = 0;
	state.busy = 1;

	state_puts ("Atari DMA sound driver:\n");

	state_puts ("\tsound.format = 0x");
	state_puti (sound.format, 16); 
	switch( sound.format ) {
	  case AFMT_MU_LAW:	state_puts( " (mu-law, interpolated)" ); break;
	  case AFMT_U8:		state_puts( " (unsigned 8 bit)" ); break;
	  case AFMT_S8:		state_puts( " (signed 8 bit)" ); break;
	}
	state_puts ("\n");
	
	state_puts ("\tsound.speed = 0x");
	state_puti (sound.speed, 16);
	switch( sound.speed ) {
	  case DMASND_MODE_6KHZ:	state_puts( " (6.26 kHz)" ); break;
	  case DMASND_MODE_12KHZ:	state_puts( " (12.5 kHz)" ); break;
	  case DMASND_MODE_25KHZ:	state_puts( " (25.0 kHz)" ); break;
	  case DMASND_MODE_50KHZ:	state_puts( " (50.1 kHz)" ); break;
	}
	state_puts ("\n");

	state_puts ("\tsound.stereo = 0x");
	state_puti (sound.stereo, 16);
	state_puts (sound.stereo == DMASND_MODE_STEREO ? " (stereo)\n" :
		    " (mono)\n" );

	state_puts ("\tsound.volume_left = ");
	state_puti (sound.volume_left, 10); state_puts (" db [-40...0]\n");

	state_puts ("\tsound.volume_right = ");
	state_puti (sound.volume_right, 10); state_puts (" db [-40...0]\n");

	state_puts ("\tsound.bass = ");
	state_puti (sound.bass, 10); state_puts (" db [-12...+12]\n");

	state_puts ("\tsound.treble = ");
	state_puti (sound.treble, 10); state_puts (" db [-12...+12]\n");

	state_puts( "\tsq.block_size = ");
	state_puti( sq.block_size, 10 );
	state_puts( " sq.max_count = ");
	state_puti( sq.max_count, 10 );
	state_puts( "\n" );

	DISABLE_IRQ();
	count     = sq.count;
	rear_size = sq.rear_size;
	playing   = sq.playing;
	syncing   = sq.syncing;
	ENABLE_IRQ();

	state_puts( "\tsq.count = ");
	state_puti( count, 10 );
	state_puts( " sq.rear_size = ");
	state_puti( rear_size, 10 );
	state_puts( "\n" );

	state_puts( "\tsq.playing = ");
	state_puti( playing, 10 );
	state_puts( " sq.syncing = ");
	state_puti( syncing, 10 );
	state_puts( "\n" );

	return 0;
}


static int state_release (void)
{
	state.busy = 0;
	return 0;
}


static int state_read (char *dest, int count)
{
	int n = state.len - state.ptr;
	if (n > count) n = count;
	if (n <= 0) return 0;
	memcpy_tofs (dest, &state.buf [state.ptr], n);
	state.ptr += n;
	return n;
}


/*
high level sound stuff
*/


static int
sound_open (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev) & 0x0f;

	switch (dev) {
	case SND_DEV_STATUS: return state_open (file->f_flags);
	case SND_DEV_CTL: return mixer_open (file->f_flags);
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: {
	    	int rc = sq_open (file->f_flags);
	    	if (rc == 0) {
			sound.format = (dev == SND_DEV_AUDIO) ? AFMT_MU_LAW : AFMT_U8;
			sound_init ();
			if (dev == SND_DEV_AUDIO)
				sound_set_speed(12517);
		}
		return rc;
	}
	default: return -ENXIO;
	}

}


static void
sound_release (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: state_release (); return;
	case SND_DEV_CTL: mixer_release (); return;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: sq_release (); return;
	default: printk ("sound_release: Unknown minor device %d\n", dev);
	}
}


static int
sound_lseek (struct inode *inode, struct file *file, off_t offset, int orig)
{
	return -EPERM;
}


static int
sound_read (struct inode *inode, struct file *file, char *buf, int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: return state_read (buf, count);
	case SND_DEV_CTL:
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return -EPERM;
	default:
		printk ("sound_read: Unknown minor device %d\n", dev);
		return -EINVAL;
	}
}


static int
sound_write (struct inode *inode, struct file *file, char *buf, int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS:
	case SND_DEV_CTL: return -EPERM;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return sq_write(buf, count);
	default:
		printk ("sound_write: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

}


static int
sound_select (struct inode *inode, struct file *file, int sel_type,
	      select_table *wait)
{
	return -EPERM;
}


static int ioctl_return (int *addr, int value)
{
	int error;

	if (value < 0) return value;

	error = verify_area (VERIFY_WRITE, addr, sizeof (int));
	if (error) return error;

	put_fs_long (value, (long*)&((addr)[0]));
	return 0;
}


static int
sound_ioctl (struct inode *inode, struct file *file,
             unsigned int cmd, unsigned long arg)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {

	case SND_DEV_STATUS: return -EPERM;
	case SND_DEV_CTL: return mixer_ioctl (inode, file, cmd, arg);

	case SND_DEV_AUDIO:
	case SND_DEV_DSP:
		switch (cmd) {

		case SNDCTL_DSP_RESET: sq_reset (); return 0;

		case SNDCTL_DSP_POST:
		case SNDCTL_DSP_SYNC: return sq_sync ();

		case SNDCTL_DSP_SPEED:
			return IOCTL_OUT (arg, sound_set_speed (IOCTL_IN (arg)));
		case SNDCTL_DSP_STEREO:
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg)));
		case SOUND_PCM_WRITE_CHANNELS:
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg) - 1) + 1);

		case SNDCTL_DSP_SETFMT:
			return IOCTL_OUT (arg, sound_set_format (IOCTL_IN (arg)));
		case SNDCTL_DSP_GETFMTS:
			return IOCTL_OUT (arg, AFMT_MU_LAW | AFMT_U8 | AFMT_S8);

		case SNDCTL_DSP_GETBLKSIZE:
			return IOCTL_OUT (arg, 10240);
		case SNDCTL_DSP_SUBDIVIDE:
		case SNDCTL_DSP_SETFRAGMENT: break;

		default: return mixer_ioctl (inode, file, cmd, arg);
		}
		break;

	default:
		printk ("sound_ioctl: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

	return( -EINVAL );
}




static struct file_operations sound_fops =
{
	sound_lseek,
	sound_read,
	sound_write,
	NULL,			/* readdir */
	sound_select,
	sound_ioctl,
	NULL,			/* mmap */
	sound_open,
	sound_release,
	NULL			/* fsync */
};


long
soundcard_init (long mem_start)
{
	struct sound_buffer buf;
	
	if (ATARIHW_PRESENT( PCM_8BIT ) &&
	    ATARIHW_PRESENT( MICROWIRE ))
	{
		if ((mfp.int_en_a & mfp.int_mk_a & 0x20) ==0 )
		{
			/* Register driver with the VFS. */
			register_chrdev (SOUND_MAJOR, "sound", &sound_fops);
			printk ("Atari DMA sound driver installed.\n");

			/* Set up sound queue, /dev/audio and /dev/dsp.  TO DO:
			   Allocate buffer space dynamically.  But I do not see
			   how to do this, since kmalloc will not allocate more
			   than 4K and vmalloc does not guarantee memory to be
			   physically contiguous, which is necessary for DMA
			   transfer. */
			/* ++roman: It would be veeery difficult to allocate
			   these buffers dynamically...  Need not only
			   contigous pages, but also in ST-Ram...

			   ++TeSche: There's no physical need to have
			   subsequent buffers on contiguous pages (would just
			   require a different buffer adress handling) so this
			   could be dealt with, but the ST-RAM really is a
			   problem... */
			buf.start = (char *) atari_stram_alloc (
				SOUND_BUFFER_SIZE, &mem_start);
			buf.size = SOUND_BUFFER_SIZE;
			sq_init (&buf, QUEUE_BLOCK_SIZE);
        
			/* Set up /dev/sndstat. */
			state_init ();

			/* Set up /dev/mixer. */
			mixer_init ();

			/* Set up timer A.  Timer A will receive a signal upon
			   end of playing from the sound hardware.  Furthermore
			   Timer A is able to count events and will cause an
			   interrupt after a programmed number of events.  So
			   all we need to keep the music playing is to provide
			   the sound hardware with new data upon an interrupt
			   from timer A. */
			mfp.tim_ct_a = 0;	/* ++roman: Stop timer before programming! */
			mfp.tim_dt_a = 1;	/* Cause interrupt after first event. */
			mfp.tim_ct_a = 8;	/* Turn on event counting. */
			mach_add_isr (IRQ_MFP_TIMA, (isrfunc)sq_interrupt, 0, NULL);
						/* Register interrupt handler. */
			mfp.int_en_a |= 0x20;	/* Turn interrupt on. */
			mfp.int_mk_a |= 0x20;
		} else
			printk ("Atari DMA sound driver: Timer A interrupt already in use\n");
	}

	return mem_start;
}


void
sound_setup (char *str, int *ints)
{
        /* check the bootstrap parameter for sound= */
	/* at the moment there are no parameters to configure */
}

#endif /* CONFIG_ATARI_DMA_SOUND */
