/*
 * ints.c -- 680x0 Linux general interrupt handling code
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/kernel_stat.h>
#include <linux/errno.h>

#include <asm/system.h>
#include <asm/irq.h>
#include <asm/page.h>

/* list is accessed 0-6 for IRQs 1-7 */
static isr_node_t *isr_list[7];

/* The number of spurious interrupts */
volatile unsigned long num_spurious;

unsigned long interrupt_stack[PAGE_SIZE/sizeof(long)];

/*
 * void init_IRQ(void)
 *
 * Parameters:	None
 *
 * Returns:	Nothing
 *
 * This function should be called during kernel startup to initialize
 * the IRQ handling routines.
 */

void init_IRQ(void)
{
    /* Setup interrupt stack pointer */
    asm ("movec %0,%/isp"
	 : : "r" (interrupt_stack + sizeof (interrupt_stack) / sizeof (long)));

    mach_init_INTS ();
}

void insert_isr (isr_node_t **listp, isr_node_t *node)
{
    unsigned long spl;
    isr_node_t *cur, *last = NULL;

    save_flags(spl);
    cli();

    cur = *listp;

    if (!cur || cur->pri > node->pri) {
	if (cur)
	    node->next = cur;
	*listp = node;
	restore_flags (spl);
	return;
    }

    while (cur->next && cur->pri <= node->pri)
    {
	last = cur;
	cur = cur->next;
    }

    if (!cur->next) {
	cur->next = node;
    } else {
	node->next = cur;
	last->next = node;
    }

    restore_flags(spl);
}

isr_node_t *new_isr_node(void)
{
    isr_node_t *np;

    static isr_node_t nodes[100];
    static int curnode = 0;

    np = &nodes[curnode++];
    if (curnode >= sizeof(nodes)/sizeof(nodes[0]))
	panic ("new_isr_node: out of nodes");

    return np;
}

int add_isr (unsigned long source, isrfunc isr, int pri, void *data)
{
    isr_node_t *p;

    if (source & IRQ_MACHSPEC)
    {
	return mach_add_isr (source, isr, pri, data);
    }

    if (source < IRQ1 || source > IRQ7)
	panic ("add_isr: Incorrect IRQ source");

    p = new_isr_node();
    p->isr = isr;
    p->pri = pri;
    p->data = data;
    p->next = NULL;

    insert_isr (&isr_list[source-1], p);

    return 1;
}

void call_isr_list(int irq, isr_node_t *p, struct intframe *fp)
{
    while (p) {
	p->isr (irq, fp, p->data);
	p = p->next;
    }
}

#undef DEBUG_LOST_INT_DECR

asmlinkage void process_int(int level, struct intframe *fp)
{
#ifdef DEBUG_LOST_INT_DECR
	int cnt = intr_count;
	static int last_spur = 0;
#endif
  	
	kstat.interrupts[level]++;  	
	call_isr_list (level, isr_list[level-1], fp);

#ifdef DEBUG_LOST_INT_DECR
	if (cnt != intr_count) {
		printk( "intr_count=%d != old_cnt=%d; resetted\n", (int)intr_count, cnt );
		intr_count = cnt;
	}
	if (num_spurious != last_spur) {
		printk( "num_spurious is now %lu\n", num_spurious );
		last_spur = num_spurious;
	}
#endif
}

int request_irq(unsigned int irq, void (*handler)(int, struct pt_regs *),
	unsigned long flags, const char * devname)
{
	if (irq == 0) { /* TIMER_IRQ */
		mach_sched_init((isrfunc)handler);
		return 0;
	}
		
	return -EINVAL;
}

void free_irq(unsigned int irq)
{
}

void enable_irq(unsigned int irq_nr)
{
}
void disable_irq(unsigned int irq_nr)
{
}
