/*
 * drivers/char/amiga_ser.c: Amiga built-in serial port driver.
 *
 * Copyright 1994 Roman Hodek, 1994 Hamish Macdonald
 *       Based on the Atari MFP driver by Roman Hodek
 * 
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */


/*
 * This file implements the driver for the Amiga built-in serial port.
 */

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/tty.h>
#include <linux/termios.h>
#include <linux/interrupt.h>
#include <linux/serial.h>
#include <linux/linkage.h>

#include <asm/bootinfo.h>
#include <asm/amigahw.h>
#include <asm/amigaints.h>
#include <asm/traps.h>
#include <asm/irq.h>

/* some serial hardware definitions */
#define SDR_OVRUN   (1<<15)
#define SDR_TBE     (1<<13)
#define SDR_TSRE    (1<<12)

#define AC_SETCLR   (1<<15)
#define AC_UARTBRK  (1<<11)

#define SER_DTR     (1<<7)
#define SER_RTS     (1<<6)
#define SER_DCD     (1<<5)
#define SER_CTS     (1<<4)
#define SER_DSR     (1<<3)

/***************************** Prototypes *****************************/

static asmlinkage void ser_rx_int( int irq, struct intframe *fp, struct async_struct *info );
static void ser_tx_int( int irq, struct intframe *fp, struct async_struct *info );
static void ser_vbl_int( int irq, struct intframe *fp, struct async_struct *info );
static void ser_init( struct async_struct *info );
static void ser_deinit( struct async_struct *info, int leave_dtr );
static void ser_enab_tx_int( struct async_struct *info, int enab_flag );
static int  ser_check_custom_divisor (struct async_struct *info,
				      int baud_base, int divisor);
static void ser_change_speed( struct async_struct *info );
static void ser_throttle( struct async_struct *info, int status );
static void ser_set_break( struct async_struct *info, int break_flag );
static void ser_get_serial_info( struct async_struct *info,
				struct serial_struct *retinfo );
static unsigned int ser_get_modem_info( struct async_struct *info );
static int ser_set_modem_info( struct async_struct *info, int new_dtr,
			      int new_rts );
static void ser_stop_receive( struct async_struct *info );
static int ser_trans_empty( struct async_struct *info );

/************************* End of Prototypes **************************/



/* SERIALSWITCH structure for the Amiga serial port
 */

static SERIALSWITCH amiga_ser_switch = {
    ser_init, ser_deinit, ser_enab_tx_int,
    ser_check_custom_divisor, ser_change_speed,
    ser_throttle, ser_set_break,
    ser_get_serial_info, ser_get_modem_info,
    ser_set_modem_info, NULL, ser_stop_receive, ser_trans_empty, NULL
};

/* Divisors for standard speeds (NTSC and PAL) */
static int baud_table[2][17] = {
    /* NTSC */
    {
	/* B0     */ 0,
	/* B50    */ 71589,
	/* B75    */ 47726,
	/* B110   */ 32540,
	/* B134	  */ 26712,
	/* B150	  */ 23862,
	/* B200	  */ 17896,
	/* B300	  */ 11930,
	/* B600	  */ 5964,
	/* B1200  */ 2892,
	/* B1800  */ 1987,
	/* B2400  */ 1490,
	/* B4800  */ 744,
	/* B9600  */ 371,
	/* B19200 */ 185,
	/* B38400 */ 92,
	/* B57600 */ 46
    },
    /* PAL */
    {
	/* B0     */ 0,
	/* B50    */ 70936,
	/* B75    */ 47290,
	/* B110   */ 32243,
	/* B134	  */ 26468,
	/* B150	  */ 23644,
	/* B200	  */ 17733,
	/* B300	  */ 11821,
	/* B600	  */ 5910,
	/* B1200  */ 2954,
	/* B1800  */ 1969,
	/* B2400  */ 1476,
	/* B4800  */ 737,
	/* B9600  */ 368,
	/* B19200 */ 183,
	/* B38400 */ 91,
	/* B57600 */ 45
    }
};
	
static __inline__ void ser_DTRoff(void)
{
    ciab.pra |= SER_DTR; /* active low */
}

static __inline__ void ser_DTRon(void)
{
    ciab.pra &= ~SER_DTR; /* active low */
}

static __inline__ void ser_RTSoff(void)
{
    ciab.pra |= SER_RTS; /* active low */
}

static __inline__ void ser_RTSon(void)
{
    ciab.pra &= ~SER_RTS; /* active low */
}

/* This variable holds the current state of the DCD/CTS bits */
static unsigned char current_ctl_bits;

static __inline__ void check_modem_status(struct async_struct *info)
{
	unsigned char bits = ciab.pra & (SER_DCD | SER_CTS);

	if (!(info->flags & ASYNC_INITIALIZED))
		return;
  
	if (bits ^ current_ctl_bits) {
		if ((bits ^ current_ctl_bits) & SER_DCD) {
			rs_dcd_changed(info, !(bits & SER_DCD));
		}

		if ((bits ^ current_ctl_bits) & SER_CTS)
			rs_check_cts(info, !(bits & SER_CTS));
	}
	current_ctl_bits = bits;
}

asmlinkage void level5handler(void);
static struct async_struct *amiga_info;
void set_evector(int vecnum, void (*handler)(void));

void amiga_serinit( struct async_struct *info )
{	
    unsigned long flags;

    save_flags (flags);
    cli();

    /* set ISRs, and then disable the rx interrupts */
    add_isr( IRQ_AMIGA_TBE, (isrfunc)ser_tx_int, 0, info);
#if 0
    add_isr( IRQ_AMIGA_RBF, (isrfunc)ser_rx_int, 0, info);
#else
    /* this doesn't really do anything, but it uses ser_rx_int and avoids
       a warning */
    add_isr( IRQ_AMIGA_RBF, (isrfunc)ser_rx_int, 0, info);
    amiga_info = info;
    /* hijack the level 5 interrupt */
    set_evector(VEC_INT5, level5handler);
#endif
    add_isr( IRQ_AMIGA_VERTB, (isrfunc)ser_vbl_int, 0, info );

    /* turn off both Tx and Rx interrupts */
    custom.intena = IF_TBE | IF_RBF;
    /* clear any pending receive interrupt */
    custom.intreq = IF_RBF;
    /* setup a transmit interrupt */
    custom.intreq = IF_SETCLR | IF_TBE;
    restore_flags (flags);

    /*
     * set the appropriate directions for the modem control flags,
     * and clear RTS and DTR
     */
    ciab.ddra |= (SER_DTR | SER_RTS);   /* outputs */
    ciab.ddra &= ~(SER_DCD | SER_CTS | SER_DSR);  /* inputs */
    
    info->type = SER_AMIGA;
    info->port = 0x100; /* !zero */
    info->sw   = &amiga_ser_switch;
}


/*
 * A new level 5 interrupt handler that doesn't mess around with
 * interrupt bottom halves.  This'll have to change when/if we
 * start using the level 5 DSKSYN interrupt for floppy access
 */

extern unsigned long intr_count;

__asm__(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(level5handler) ":
	moveml	%a0-%a1/%d0-%d1,%sp@-
	addql	#1," SYMBOL_NAME_STR(intr_count) "
	jbsr	" SYMBOL_NAME_STR(ser_rx_int) "
	subql	#1," SYMBOL_NAME_STR(intr_count) "
	moveml	%sp@+,%a0-%a1/%d0-%d1
	rte
");

static asmlinkage void ser_rx_int(int irq,
		       struct intframe *fp, struct async_struct *info )
{
	ushort ints = custom.intreqr & custom.intenar;
	int ch;

	if (ints & IF_DSKSYN) {
		printk(KERN_ERR "Amiga Floppy Disk Sync Interrupt!\n");
		custom.intreq = IF_DSKSYN;
	}

	if (!(ints & IF_RBF))
		return;

	do {
		ch = custom.serdatr;
		/* 
		 * clear the interrupt, so that another character can be read
		 */
		custom.intreq = IF_RBF;

		rs_receive_char(amiga_info, ch & 0xff, 
				ch & SDR_OVRUN ? TTY_OVERRUN : 0);
	} while (custom.intreqr & IF_RBF);

	/* get new serial info (DCD, CTS) */
	check_modem_status (amiga_info);
}


static void ser_tx_int( int irq,
		        struct intframe *fp, struct async_struct *info )
{
	/* get new serial info (DCD, CTS) */
	check_modem_status (info);

	if ((info->xmit_cnt <= 0) || info->tty->stopped ||
	    info->tty->hw_stopped) {
		/* disable tx interrupts */
		custom.intena = IF_TBE;
		return;
	}

	if (info->x_char) {
		/* clear previous interrupt */
		custom.intreq = IF_TBE;
		custom.serdat = info->x_char | 0x100;
		info->x_char = 0;
		return;
	}

	/* clear previous interrupt */
	custom.intreq = IF_TBE;
	custom.serdat = (info->xmit_buf[info->xmit_tail++]) | 0x100;
	info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
	--info->xmit_cnt;
	
	if (info->xmit_cnt < WAKEUP_CHARS)
		rs_sched_event(info, RS_EVENT_WRITE_WAKEUP);
	
	if (info->xmit_cnt <= 0)
		/* disable tx interrupts */
		custom.intena = IF_TBE;
}

static void ser_vbl_int( int irq,
			 struct intframe *fp, struct async_struct *info )
{	
	check_modem_status(info);
}

static void ser_init( struct async_struct *info )
{	
	/* enable Rx interrupts */
	custom.intena = IF_SETCLR | IF_RBF;

	/* setup a pending transmit interrupt */
	custom.intreq = IF_SETCLR | IF_TBE;

	/* turn on DTR and RTS */
	ser_DTRon();
	ser_RTSon();

	/* remember current state of the DCD and CTS bits */
	current_ctl_bits = ciab.pra & (SER_DCD | SER_CTS);
}

static void ser_enab_tx_int (struct async_struct *info, int enab_flag)
{
    if (enab_flag) {
	/* enable tx interrupts (should result in an immediate interrupt) */
	custom.intena = IF_SETCLR | IF_TBE;
    } else
	/* disable tx interrupts */
	custom.intena = IF_TBE;
}

static void ser_deinit( struct async_struct *info, int leave_dtr )
{
    /* disable Rx and Tx interrupts */
    custom.intena = IF_RBF | IF_TBE;

    /* wait for last byte to be completely shifted out */
    while( !(custom.serdatr & SDR_TSRE) )
	;

    /* drop RTS and DTR if required */
    ser_RTSoff();
    if (!leave_dtr)
	ser_DTRoff();
}

static int ser_check_custom_divisor (struct async_struct *info,
				     int baud_base, int divisor)
{
    /* allow any divisor */
    return 0;
}


static void ser_change_speed( struct async_struct *info )

{
    unsigned	cflag, baud, chsize, stopb, parity, aflags;
    unsigned	div = 0;
	
    if (!info->tty || !info->tty->termios) return;

    cflag  = info->tty->termios->c_cflag;
    baud   = cflag & CBAUD;
    chsize = cflag & CSIZE;
    stopb  = cflag & CSTOPB;
    parity = cflag & (PARENB | PARODD);
    aflags = info->flags & ASYNC_SPD_MASK;

    if (cflag & CRTSCTS)
	info->flags |= ASYNC_CTS_FLOW;
    else
	info->flags &= ~ASYNC_CTS_FLOW;
    if (cflag & CLOCAL)
	info->flags &= ~ASYNC_CHECK_CD;
    else
	info->flags |= ASYNC_CHECK_CD;

    if (baud & CBAUDEX) {
	    baud &= ~CBAUDEX;
	    if (baud < 1 || baud > 2)
		    info->tty->termios->c_cflag &= ~CBAUDEX;
	    else
		    baud += 15;
    }
    if (baud == 15) {
	    if (aflags == ASYNC_SPD_HI)
		    baud += 1;
	    if (aflags == ASYNC_SPD_VHI)
		    baud += 2;
	    if (aflags == ASYNC_SPD_CUST)
		    div = info->custom_divisor;
    }
    if (!div){
	/* Maximum  speed is 57600 :-( */
	if (baud > 16) baud = 16;
	if (boot_info.bi_amiga.eclock == NTSC_ECLOCK)
	    div = baud_table[0][baud];
	else
	    div = baud_table[1][baud];
    }

    if (div) {
	    /* turn on DTR */
	    ser_DTRon();
    } else {
	    /* speed == 0 -> drop DTR */
	    ser_DTRoff();
	    return;
    }

    /* setup the serial port period register */
    custom.serper = div;
}
	
	
static void ser_throttle( struct async_struct *info, int status )
{
    if (status)
	ser_RTSoff();
    else
	ser_RTSon();
}


static void ser_set_break( struct async_struct *info, int break_flag )
{
    if (break_flag)
	custom.adkcon = AC_SETCLR | AC_UARTBRK;
    else
	custom.adkcon = AC_UARTBRK;
}


static void ser_get_serial_info( struct async_struct *info,
				struct serial_struct *retinfo )
{
    retinfo->baud_base = 0;
    retinfo->custom_divisor = info->custom_divisor;
}


static unsigned int ser_get_modem_info( struct async_struct *info )
{
    unsigned int minfo = ciab.pra;

    return(
	   ((minfo & SER_DTR) ? 0 : TIOCM_DTR) |
	   ((minfo & SER_RTS) ? 0 : TIOCM_RTS) |
	   ((minfo & SER_DCD) ? 0 : TIOCM_CAR) |
	   ((minfo & SER_CTS) ? 0 : TIOCM_CTS) |
	   ((minfo & SER_DSR) ? 0 : TIOCM_DSR) |
	   /* TICM_RNG */ 0
	   );
}


static int ser_set_modem_info( struct async_struct *info,
			      int new_dtr, int new_rts )
{
    if (new_dtr == 0)
	ser_DTRoff();
    else if (new_dtr == 1)
	ser_DTRon();
    
    if (new_rts == 0)
	ser_RTSoff();
    else if (new_rts == 1)
	ser_RTSon();

    return( 0 );
}

static void ser_stop_receive( struct async_struct *info )
{
	/* disable receive interrupts */
	custom.intena = IF_RBF;
	/* clear any pending receive interrupt */
	custom.intreq = IF_RBF;
}

static int ser_trans_empty( struct async_struct *info )
{
       return !!(custom.serdatr & SDR_TSRE);
}
