/*
 * atari_scsi.c -- Device dependant functions for the Atari generic SCSI port
 *
 * Copyright 1994 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 *
 *   Loosely based on the work of Robert De Vries' team and added:
 *    - working real DMA
 *    - Falcon support (untested yet!)   ++bjoern fixed and now it works
 *    - lots of extensions and bug fixes.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */


/**************************************************************************/
/*                                                                        */
/* Notes for Falcon SCSI:                                                 */
/* ----------------------                                                 */
/*                                                                        */
/* Since the Falcon SCSI uses the ST-DMA chip, that is shared among       */
/* several device drivers, locking and unlocking the access to this       */
/* chip is required. But locking is not possible from an interrupt,       */
/* since it puts the process to sleep if the lock is not available.       */
/* This prevents "late" locking of the DMA chip, i.e. locking it just     */
/* before using it, since in case of disconnection-reconnection           */
/* commands, the DMA is started from the reselection interrupt.           */
/*                                                                        */
/* Two possible schemes for ST-DMA-locking would be:                      */
/*  1) The lock is taken for each command separately and disconnecting    */
/*     is forbidden (i.e. can_queue = 1).                                 */
/*  2) The DMA chip is locked when the first command comes in and         */
/*     released when the last command is finished and all queues are      */
/*     empty.                                                             */
/* The first alternative would result in bad performance, since the       */
/* interleaving of commands would not be used. The second is unfair to    */
/* other drivers using the ST-DMA, because the queues will seldom be      */
/* totally empty if there is a lot of disk traffic.                       */
/*                                                                        */
/* For this reasons I decided to employ a more elaborate scheme:          */
/*  - First, we give up the lock everytime we can (for fairness), this    */
/*    means every time a command finishes and there are no other commands */
/*    on the disconnected queue.                                          */
/*  - If there are others waiting to lock the DMA chip, we stop           */
/*    issueing commands, i.e. moving them onto the issue queue.           */
/*    Because of that, the disconnected queue will run empty in a         */
/*    while. Instead we go to sleep on a 'fairness_queue'.                */
/*  - If the lock is released, all processes waiting on the fairness      */
/*    queue will be woken. The first of them trys to re-lock the DMA,     */
/*    the others wait for the first to finish this task. After that,      */
/*    they can all run on and do their commands...                        */
/* This sounds complicated (and it is it :-(), but it seems to be a       */
/* good compromise between fairness and performance: As long as noone     */
/* else wants to work with the ST-DMA chip, SCSI can go along as          */
/* usual. If now someone else comes, this behaviour is changed to a       */
/* "fairness mode": just already initiated commands are finished and      */
/* then the lock is released. The other one waiting will probably win     */
/* the race for locking the DMA, since it was waiting for longer. And     */
/* after it has finished, SCSI can go ahead again. Finally: I hope I      */
/* have not produced any deadlock possibilities!                          */
/*                                                                        */
/**************************************************************************/



#include <linux/config.h>

#ifdef CONFIG_ATARI_SCSI

/* #define NDEBUG (NDEBUG_DMA) */

#define AUTOSENSE
/* For the Atari version, use only polled IO or REAL_DMA */
#define	REAL_DMA

#include <linux/types.h>
#include <linux/stddef.h>
#include <linux/delay.h>
#include <linux/mm.h>

#include <asm/bootinfo.h>
#include <asm/atarihw.h>
#include <asm/atariints.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/irq.h>

#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "atari_scsi.h"
#include "NCR5380.h"
#include "constants.h"
#include <asm/atari_stdma.h>
#include <asm/io.h>


#define	IS_A_TT()	ATARIHW_PRESENT(TT_SCSI)

#define	SCSI_DMA_WRITE_P(elt,val)				\
	do {										\
		unsigned long v = val;					\
		tt_scsi_dma.elt##_lo = v & 0xff;		\
		v >>= 8;								\
		tt_scsi_dma.elt##_lmd = v & 0xff;		\
		v >>= 8;								\
		tt_scsi_dma.elt##_hmd = v & 0xff;		\
		v >>= 8;								\
		tt_scsi_dma.elt##_hi = v & 0xff;		\
	} while(0)

#define	SCSI_DMA_READ_P(elt)							\
	(((unsigned long)tt_scsi_dma.elt##_hi  << 24) |		\
	 ((unsigned long)tt_scsi_dma.elt##_hmd << 16) |		\
	 ((unsigned long)tt_scsi_dma.elt##_lmd << 8) |		\
	  (unsigned long)tt_scsi_dma.elt##_lo)


#define	SCSI_DMA_SETADR(adr)					\
    do {										\
		unsigned long __adr = (adr);			\
		st_dma.dma_lo = (unsigned char)__adr;	\
		__adr >>= 8;							\
		st_dma.dma_md = (unsigned char)__adr;	\
		__adr >>= 8;							\
		st_dma.dma_hi = (unsigned char)__adr;	\
	} while(0)

#define	SCSI_DMA_GETADR() ({					\
    unsigned long	__adr;						\
	__adr = st_dma.dma_lo;						\
    __asm__ __volatile__ ("nop");				\
    __adr |= (st_dma.dma_md & 0xff) << 8;		\
	__adr |= (st_dma.dma_hi & 0xff) << 16;		\
	__adr;										\
})



/***************************** Prototypes *****************************/

#ifdef REAL_DMA
static void scsi_dma_buserr( int irq, struct intframe *fp, void *data);
static void atari_scsi_fetch_restbytes(void);
static long atari_scsi_dma_residual(struct Scsi_Host *instance);
static unsigned long atari_dma_xfer_len(unsigned long wanted_len);
#endif
static void scsi_tt_intr( int irq, struct intframe *fp, void *data);
static void scsi_falcon_intr( int irq, struct intframe *fp, void *data);
static void falcon_release_lock_if_possible( struct NCR5380_hostdata *
                                             hostdata );
static void falcon_get_lock( void );
static unsigned char atari_scsi_tt_reg_read( unsigned char reg );
static void atari_scsi_tt_reg_write( unsigned char reg, unsigned char value
                                     );
static unsigned char atari_scsi_falcon_reg_read( unsigned char reg );
static void atari_scsi_falcon_reg_write( unsigned char reg, unsigned char
                                         value );

/************************* End of Prototypes **************************/


static struct Scsi_Host *atari_scsi_host = NULL;
static unsigned char (*atari_scsi_reg_read)( unsigned char reg );
static void (*atari_scsi_reg_write)( unsigned char reg, unsigned char value );

#ifdef REAL_DMA
static unsigned long	atari_dma_residual, atari_dma_startaddr;
static short			atari_dma_active;
#endif


#if defined(REAL_DMA)

static int scsi_dma_is_ignored_buserr( unsigned char dma_stat )

{	int				i;
	unsigned long	addr = SCSI_DMA_READ_P( dma_addr ), end_addr;

	if (dma_stat & 0x01) {

		/* A bus error happens when DMA-ing from the last page of a
		 * physical memory chunk (DMA prefetch!), but that doesn't hurt.
		 * Check for this case:
		 */
		
		for( i = 0; i < boot_info.num_memory; ++i ) {
			end_addr = boot_info.memory[i].addr +
				boot_info.memory[i].size;
			if (end_addr <= addr && addr <= end_addr + 4)
				return( 1 );
		}
	}
	return( 0 );
}


static void scsi_dma_buserr (int irq, struct intframe *fp, void *data)

{	unsigned char	dma_stat = tt_scsi_dma.dma_ctrl;

	printk("Bad SCSI DMA interrupt! dma_addr=0x%08lx dma_stat=%02x dma_cnt=%08lx\n",
	       SCSI_DMA_READ_P(dma_addr), dma_stat, SCSI_DMA_READ_P(dma_cnt));
	if (dma_stat & 0x80) {
		if (!scsi_dma_is_ignored_buserr( dma_stat ))
			panic( "SCSI DMA bus error -- bad DMA programming!" );
	}
	else {
		/* Under normal circumstances we never should get to this point,
		 * since both interrupts are triggered simultaneously and the 5380
		 * int has more priority. When this irq is handled, that DMA
		 * interrupt is cleared. So a warning message is printed here.
		 */
		printk( "SCSI DMA intr ?? -- this shouldn't happen!\n" );
	}
}

#endif


static void scsi_tt_intr (int irq, struct intframe *fp, void *data)
{
	unsigned long	flags;
#ifdef REAL_DMA
	int dma_stat;

	dma_stat = tt_scsi_dma.dma_ctrl;

#if (NDEBUG & NDEBUG_INTR)
	printk("scsi%d: NCR5380 interrupt, DMA status = %02x\n",
	       atari_scsi_host->host_no, dma_stat & 0xff);
#endif

	/* Look if it was the DMA that has interrupted: First possibility
	 * is that a bus error occured...
	 */
	if (dma_stat & 0x80) {
		if (!scsi_dma_is_ignored_buserr( dma_stat )) {
			printk( "SCSI DMA caused bus error near 0x%08lx\n",
					SCSI_DMA_READ_P( dma_addr ));
			panic( "SCSI DMA bus error -- bad DMA programming!" );
		}
	}

	/* If the DMA is active but not finished, we have the the case
	 * that some other 5380 interrupt occured within the DMA transfer.
	 * This means we have residual bytes, if the desired end address
	 * is not yet reached. Maybe we have to fetch some bytes from the
	 * rest data register, too. The residual must be calculated from
	 * the address pointer, not the counter register, because only the
	 * addr reg counts bytes not yet written and pending in the rest
	 * data reg!
	 */
	if ((dma_stat & 0x02) && !(dma_stat & 0x40)) {
		atari_dma_residual =
			((struct NCR5380_hostdata *)(atari_scsi_host->hostdata))->dma_len -
				(SCSI_DMA_READ_P( dma_addr ) - atari_dma_startaddr);
#if (NDEBUG & NDEBUG_DMA)
		printk( "SCSI DMA: There are %ld residual bytes.\n",
				atari_dma_residual );
#endif
		if ((signed int)atari_dma_residual < 0)
			atari_dma_residual = 0;
		if ((dma_stat & 1) == 0) {
			/* After read operations, we maybe have to
			   transport some rest bytes */
			atari_scsi_fetch_restbytes();
		}
		else {
			/* There seems to be a nasty bug in some SCSI-DMA/NCR
			   combinations: If a target disconnects while a write
			   operation is going on, the address register of the
			   DMA may be a few bytes farer than it actually read.
			   This is probably due to DMA prefetching and a delay
			   between DMA and NCR.  Experiments showed that the
			   dma_addr is 9 bytes to high, but this could vary.
			   The problem is, that the residual is thus calculated
			   wrong and the next transfer will start behind where
			   it should.  So we round up the residual to the next
			   multiple of a sector size, if it isn't already a
			   multiple and the originally expected transfer size
			   was.  The latter condition is there to ensure that
			   the correction is taken only for "real" data
			   transfers and not for, e.g., the parameters of some
			   other command.  These shouldn't disconnect anyway.
			   */
			if (atari_dma_residual & 0x1ff) {
#if (NDEBUG & NDEBUG_DMA)
				printk("SCSI DMA: DMA bug corrected, difference %ld bytes\n",
				       512 - (atari_dma_residual & 0x1ff));
#endif
				atari_dma_residual = (atari_dma_residual + 511) & ~0x1ff;
			}
		}
		tt_scsi_dma.dma_ctrl = 0;
	}

	/* If the DMA is finished, fetch the rest bytes and turn it off */
	if (dma_stat & 0x40) {
		atari_dma_residual = 0;
		if ((dma_stat & 1) == 0)
			atari_scsi_fetch_restbytes();
		tt_scsi_dma.dma_ctrl = 0;
	}

#endif /* REAL_DMA */
	
	/* If we got this interrupt, we don't need the other one any more */
	tt_mfp.int_pn_b = (unsigned char)~0x80;

	save_flags(flags);
	NCR5380_intr (0, 0);
	restore_flags(flags);
}


static void scsi_falcon_intr (int irq, struct intframe *fp, void *data)
{
	unsigned long	flags;
#ifdef REAL_DMA
	int dma_stat;

	/* Turn off DMA and select sector counter register before
	 * accessing the status register (Atari recommendation!)
	 */
	st_dma.dma_mode_status = 0x90;
	dma_stat = st_dma.dma_mode_status;

	/* Bit 0 indicates some error in the DMA process... don't know
	 * what happened exactly (no further docu).
	 */
	if (!(dma_stat & 0x01)) {
		/* DMA error */
		printk( "SCSI DMA error near 0x%08lx!\n", SCSI_DMA_GETADR() );
	}

	/* If the DMA was active, but now bit 1 is not clear, it is some
	 * other 5380 interrupt that finishes the DMA transfer. We have to
	 * calculate the number of residual bytes and give a warning if
	 * bytes are stuck in the ST-DMA fifo (there's no way to reach them!)
	 */
	if (atari_dma_active && (dma_stat & 0x02)) {
		unsigned long	transferred;

		transferred = SCSI_DMA_GETADR() - atari_dma_startaddr;
		/* The ST-DMA address is incremented in 2-byte steps, but the
		 * data are written only in 16-byte chunks. If the number of
		 * transferred bytes is not divisible by 16, the remainder is
		 * lost somewhere in outer space.
		 */
		if (transferred & 15)
			printk( "SCSI DMA error: %ld bytes lost in ST-DMA fifo :-((\n",
					transferred & 15 );

		atari_dma_residual =
			((struct NCR5380_hostdata *)(atari_scsi_host->hostdata))->dma_len -
				transferred;
#if (NDEBUG & NDEBUG_DMA)
		printk( "SCSI DMA: There are %ld residual bytes.\n",
				atari_dma_residual );
#endif
	}
	else
		atari_dma_residual = 0;
	atari_dma_active = 0;
	
#endif /* REAL_DMA */

	save_flags(flags);
	NCR5380_intr (0, 0);
	restore_flags(flags);
}


#ifdef REAL_DMA
static void atari_scsi_fetch_restbytes( void )

{	int		nr;
	char	*src, *dst;

	/* fetch rest bytes in the DMA register */
	dst = (char *)SCSI_DMA_READ_P( dma_addr );
	if ((nr = ((long)dst & 3))) {
		/* there are 'nr' bytes left for the last long address before the
		   DMA pointer */
		dst = (char *)( (unsigned long)dst & ~3 );
#if (NDEBUG & NDEBUG_DMA)
		printk( "SCSI DMA: there are %d rest bytes for phys addr 0x%08lx",
				nr, (long)dst );
#endif
		dst = (char *)PTOV(dst);  /* The content of the DMA pointer
								   * is a physical address! */
#if (NDEBUG & NDEBUG_DMA)
		printk( " = virt addr 0x%08lx\n", (long)dst );
#endif
		for( src = (char *)&tt_scsi_dma.dma_restdata; nr > 0; --nr )
			*dst++ = *src++;
	}
}
#endif /* REAL_DMA */


static int					falcon_got_lock = 0;
static struct wait_queue	*falcon_fairness_wait = NULL;
static int					falcon_trying_lock = 0;
static struct wait_queue	*falcon_try_wait = NULL;
static int			falcon_dont_release = 0;

/* This function releases the lock on the DMA chip if there is no
 * connected command and the disconnected queue is empty. On
 * releasing, instances of falcon_get_lock are awoken, that put
 * themselves to sleep for fairness. They can now try to get the lock
 * again (but others waiting longer more probably will win).
 */

static void
falcon_release_lock_if_possible( struct NCR5380_hostdata * hostdata )

{
	unsigned long	oldflags;
		
	if (IS_A_TT()) return;
	
	save_flags(oldflags);
	cli();

	if (falcon_got_lock &&
		!hostdata->disconnected_queue &&
		!hostdata->issue_queue &&
		!hostdata->connected) {

		if (falcon_dont_release) {
#if 0
			printk("WARNING: Lock release not allowed. Ignored\n");
#endif
			restore_flags(oldflags);
			return;
		}
		falcon_got_lock = 0;
		stdma_release();
		wake_up( &falcon_fairness_wait );
	}

	restore_flags(oldflags);
}

/* This function manages the locking of the ST-DMA.
 * If the DMA isn't locked already for SCSI, it trys to lock it by
 * calling stdma_lock(). But if the DMA is locked by the SCSI code and
 * there are other drivers waiting for the chip, we do not issue the
 * command immediately but wait on 'falcon_fairness_queue'. We will be
 * waked up when the DMA is unlocked by some SCSI interrupt. After that
 * we try to get the lock again.
 * But we must be prepared that more then one instance of
 * falcon_get_lock() is waiting on the fairness queue. They should not
 * try all at once to call stdma_lock(), one is enough! For that, the
 * first one sets 'falcon_trying_lock', others that see that variable
 * set wait on the queue 'falcon_try_wait'.
 * Complicated, complicated.... Sigh...
 */

static void falcon_get_lock( void )

{
	unsigned long	oldflags;

	if (IS_A_TT()) return;

	save_flags(oldflags);
	cli();

	while( intr_count == 0 && falcon_got_lock && stdma_others_waiting() )
		sleep_on( &falcon_fairness_wait );

	while (!falcon_got_lock) {
		if (intr_count > 0)
			panic( "Falcon SCSI hasn't ST-DMA lock in interrupt" );
		if (!falcon_trying_lock) {
			falcon_trying_lock = 1;
			stdma_lock( scsi_falcon_intr, 0 );
			falcon_got_lock = 1;
			falcon_trying_lock = 0;
			wake_up( &falcon_try_wait );
		}
		else {
			sleep_on( &falcon_try_wait );
		}
	}	

	restore_flags(oldflags);
	if (!falcon_got_lock)
		panic("Falcon SCSI: someone stole the lock :-(\n");
}


/* This is the wrapper function for NCR5380_queue_command(). It just
 * trys to get the lock on the ST-DMA (see above) and then calls the
 * original function.
 */

#if 0
int atari_queue_command (Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *))
{
	/* falcon_get_lock();
	 * ++guenther: moved to NCR5380_queue_command() to prevent
	 * race condition, see there for an explanation.
	 */
	return( NCR5380_queue_command( cmd, done ) );
}
#endif


#define	RTC_READ(reg)							\
    ({	unsigned char	__val;					\
		outb(reg,&tt_rtc.regsel);				\
		__val = tt_rtc.data;					\
		__val;									\
	})

#define	RTC_WRITE(reg,val)						\
    do {										\
		outb(reg,&tt_rtc.regsel);				\
		tt_rtc.data = (val);					\
	} while(0)
    
				   
int atari_scsi_detect (Scsi_Host_Template *host)

{	static int called = 0;
	struct Scsi_Host *instance;

	if (boot_info.machtype != MACH_ATARI ||
	    (!ATARIHW_PRESENT(ST_SCSI) && !ATARIHW_PRESENT(TT_SCSI)) ||
	    called)
		return( 0 );

	atari_scsi_reg_read  = IS_A_TT() ? atari_scsi_tt_reg_read :
									   atari_scsi_falcon_reg_read;
	atari_scsi_reg_write = IS_A_TT() ? atari_scsi_tt_reg_write :
									   atari_scsi_falcon_reg_write;

	instance = scsi_register (host, sizeof (struct NCR5380_hostdata));
	atari_scsi_host = instance;

	/* Test if a host id is set in the NVRam */
	if (ATARIHW_PRESENT(TT_CLK)) {
		unsigned char sum = 0, b;
		int i;

		/* Make checksum */
		for( i = 14; i < 62; ++i )
			sum += RTC_READ(i);

		if (/* NV-Ram checksum valid? */
			RTC_READ(62) == sum && RTC_READ(63) == ~sum &&
			/* Arbitration enabled? (for TOS) */
			(b = RTC_READ( 30 )) & 0x80) {
			atari_scsi_host->this_id =
			atari_scsi_host->hostt->this_id = b & 7;
		}
	}

	NCR5380_init (instance, 0);

	if (IS_A_TT()) {
		
		add_isr( IRQ_TT_MFP_SCSI, scsi_tt_intr, 0, NULL);
		tt_mfp.active_edge |= 0x80;		/* SCSI int on L->H */
		tt_mfp.int_en_a |= 0x80;		/* enabled */
		tt_mfp.int_mk_a |= 0x80;		/* not masked */
#ifdef REAL_DMA
		/* On the TT, we got a second interrupt for DMA ready and DMA buserror.
		 * Since on DMA ready we get a "normal" interrupt, too, the service
		 * routine for the second int just checks for buserrs.
		 */
		add_isr( IRQ_TT_MFP_SCSIDMA, scsi_dma_buserr, 0, NULL);
		tt_mfp.active_edge &= ~0x20;	/* DMA int on H->L */
		tt_mfp.int_en_b |= 0x80;		/* enabled */
		tt_mfp.int_mk_b |= 0x80;		/* not masked */

		tt_scsi_dma.dma_ctrl = 0;
		atari_dma_residual = 0;
#endif /* REAL_DMA */

	}
	else { /* ! IS_A_TT */
		
		/* Nothing to do for the interrupt: the ST-DMA is initialized
		 * already by atari_init_INTS()
		 */

#ifdef REAL_DMA
		atari_dma_residual = 0;
		atari_dma_active = 0;
#endif

		/* Falcon's ST-DMA can work only in lower 16 MB, so
		 * unchecked_isa_dma has to be set. But... the high level
		 * codes allocates lots (!) of memory for buffers if this is
		 * done. So the values for sg_tablesize and cmd_per_lun get
		 * lowered a bit (you can leave this out if you have 16MB of
		 * ST-Ram).
		 */
		atari_scsi_host->unchecked_isa_dma =
		atari_scsi_host->hostt->unchecked_isa_dma = 1;
		atari_scsi_host->sg_tablesize =
		atari_scsi_host->hostt->sg_tablesize = SG_NONE;
		atari_scsi_host->hostt->cmd_per_lun = 1;
		atari_scsi_host->hostt->can_queue = 1;
	}

	printk( "scsi%d: options CAN_QUEUE=%d CMD_PER_LUN=%d SCATTERGATHER=%d HOSTID=%d",
			instance->host_no, instance->hostt->can_queue,
			instance->hostt->cmd_per_lun,
			instance->hostt->sg_tablesize,
			instance->hostt->this_id );
	NCR5380_print_options (instance);
	printk ("\n");

	called = 1;

	return( 1 );
}

int atari_scsi_reset( Scsi_Cmnd *cmd )
{
	int		rv;
	struct NCR5380_hostdata *hostdata =
		(struct NCR5380_hostdata *)cmd->host->hostdata;

	/* For doing the reset, SCSI interrupts must be disabled first,
	 * since the 5380 raises its IRQ line while _RST is active and we
	 * can't disable interrupts completely, since we need the timer.
	 */
	/* And abort a maybe active DMA transfer */
	if (IS_A_TT()) {
		tt_mfp.int_en_a &= ~0x80;
#ifdef REAL_DMA
		tt_scsi_dma.dma_ctrl = 0;
#endif /* REAL_DMA */
	}
	else {
		mfp.int_en_b &= ~0x80;
#ifdef REAL_DMA
		st_dma.dma_mode_status = 0x90;
		atari_dma_active = 0;
#endif /* REAL_DMA */
	}

	rv = NCR5380_reset( cmd );

	/* Re-enable ints */
	if (IS_A_TT()) {
		tt_mfp.int_en_a |= 0x80;
	}
	else {
		mfp.int_en_b |= 0x80;
	}
	falcon_release_lock_if_possible(hostdata);

	return( rv );
}

	
const char * atari_scsi_info (struct Scsi_Host *host)

{	/* atari_scsi_detect() is verbose enough... */
	static const char string[] = "";
	return string;
}


#if defined(REAL_DMA)

unsigned long atari_scsi_dma_setup( struct Scsi_Host *instance, void *data,
									unsigned long count, int dir )
{
	unsigned long addr = VTOP( data ),
	              cando = atari_dma_xfer_len( count );
	extern int is_medusa;

#if (NDEBUG & NDEBUG_DMA)
	printk ("scsi%d: setting up dma, data = %p, phys = %lx, count = %ld, cando = %ld, dir = %d\n",
			instance->host_no, data, addr, count, cando, dir);
#endif

	atari_dma_startaddr = addr;	/* Needed for calculating residual later. */
  
	/* Cache cleanup stuff: On writes, push any dirty cache out before sending
	 * it to the peripheral. (Must be done before DMA setup, since at least
	 * the ST-DMA begins to fill internal buffers right after setup. For
	 * reads, invalidate any cache, may be altered after DMA without CPU
	 * knowledge.
	 * 
	 * ++roman: For the Medusa, there's no need at all for that cache stuff,
	 * because the hardware does bus snooping (fine!).
	 */
	if (!is_medusa) {
		if (dir)
			cache_push( addr, cando );
		else
			cache_clear( addr, cando );
	}

	if (cando == 0)
		printk("SCSI warning: DMA programmed for 0 bytes !\n");

	if (IS_A_TT()) {
		tt_scsi_dma.dma_ctrl = dir;
		SCSI_DMA_WRITE_P( dma_addr, addr );
		SCSI_DMA_WRITE_P( dma_cnt, cando );
		tt_scsi_dma.dma_ctrl = dir | 2;
	}
	else { /* ! IS_A_TT */
  
		/* set address */
		SCSI_DMA_SETADR( addr );

		/* toggle direction bit to clear FIFO and set DMA direction */
		dir <<= 8;
		st_dma.dma_mode_status = 0x90 | dir;
		st_dma.dma_mode_status = 0x90 | (dir ^ 0x100);
		st_dma.dma_mode_status = 0x90 | dir;
		udelay(40);
		st_dma.fdc_acces_seccount = cando >> 9;
		udelay(40);
		st_dma.dma_mode_status = 0x10 | dir;
		udelay(40);
		/* need not restore value of dir, only boolean value is tested */
		atari_dma_active = 1;
	}

	return( cando );
}


static long atari_scsi_dma_residual( struct Scsi_Host *instance )

{
	return( atari_dma_residual );
}

/* This function calculates the number of bytes that can be transferred
 * via DMA. On the TT, this is arbitrary, but on the Falcon we have to use
 * the ST-DMA chip. There are only multiples of 512 bytes possible and
 * max. 256*512 bytes :-( This means also, that defining READ_OVERRUNS is
 * not possible, since that would require to program the DMA for n*512 - 2
 * bytes.
 */

static unsigned long atari_dma_xfer_len( unsigned long wanted_len )

{
	unsigned long	possible_len;

	if (IS_A_TT())
		/* TT SCSI DMA can transfer arbitrary #bytes */
		return( wanted_len );

	/* ST DMA chip is stupid -- only multiples of 512 bytes! (and max.
	 * 255*512 bytes, but this should be enough)
	 *
	 * ++roman: Some drivers (specially st.c) start commands that send only a
	 * few bytes with a transfer size of 512 (as a maximum). If a possible_len
	 * > 0 is returned here, DMA will be used and most propably some bytes
	 * will get stuck in the ST-DMA-FIFO. So let's pretend DMA is possible
	 * only from 1024 bytes on, which is the usual min. size for a real block
	 * transfer.
	 */
	possible_len =wanted_len & ~0x1ff;
	if (possible_len == 512)
		possible_len = 0;
	if (possible_len > 255*512)
		possible_len = 255*512;

#if (NDEBUG & NDEBUG_DMA)
	if (possible_len != wanted_len)
		printk( "Sorry, must cut DMA transfer size to %ld bytes instead of %ld\n",
				possible_len, wanted_len );
#endif

	return( possible_len );
}


#endif	/* REAL_DMA */


/* NCR5380 register access functions
 *
 * There are separate functions for TT and Falcon, because the access
 * methods are quite different. The calling macros NCR5380_read and
 * NCR5380_write call these functions via function pointers.
 */

static unsigned char atari_scsi_tt_reg_read( unsigned char reg )

{
	return( tt_scsi_regp[reg * 2] );
}

static void atari_scsi_tt_reg_write( unsigned char reg, unsigned char value )

{
	tt_scsi_regp[reg * 2] = value;
}

static unsigned char atari_scsi_falcon_reg_read( unsigned char reg )

{
	dma_wd.dma_mode_status= (u_short)(0x88 + reg);
    return( (u_char)dma_wd.fdc_acces_seccount );
}

static void atari_scsi_falcon_reg_write( unsigned char reg, unsigned char value )

{
	dma_wd.dma_mode_status = (u_short)(0x88 + reg);
	dma_wd.fdc_acces_seccount = (u_short)value;
}

static inline void acknowledge_scsi_int (void)
{
  if (IS_A_TT ())
    tt_mfp.int_sv_a = (char) ~0x80;
  else
    mfp.int_sv_b = (char) ~0x80;
}

#include "atari_NCR5380.c"

#endif /* CONFIG_ATARI_*_SCSI */
