/*
 *	mvme166_scsi.c  low level driver for ncr 53c710 on mvme166_scsi brd
 *	Copyright (C) 1995 R.G.Hirst, 01 Oct 95
 *
 *  based on:
 *
 *	seagate.c Copyright (C) 1992 Drew Eckhardt 
 *	low level scsi driver for ST01/ST02 by
 *		Drew Eckhardt 
 *
 *	<drew@colorado.edu>
 */

#include <asm/io.h>
#include <asm/system.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/config.h>

#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "mvme166_scsi.h"
#include "constants.h"

#define xxDEBUG (PRINT_COMMAND | PHASE_EXIT)
#define xDEBUG 0xffff

static int internal_command(unsigned char target, unsigned char lun,
			    const void *cmnd,
			 void *buff, int bufflen);

static int incommand;			/*
						set if arbitration has finished
						and we are 
						in some command phase.
					*/

static volatile int abort_confirm = 0;

static volatile NCR710 *ncr710;			/* SCSI chip pointer */

static volatile int ncr_aborted=0;	/* 
						set when we are aborted, ie by
						a time out, etc.
					*/

#define retcode(result) (((result) << 16) | (message << 8) | status) 			

/*
 * hostno stores the hostnumber, as told to us by the init routine.
 */

static int hostno = -1;


int mvme166_scsi_detect (Scsi_Host_Template * tpnt)
{
	struct Scsi_Host *instance;
	unsigned clock;
#ifdef DEBUG 
	printk("Autodetecting MVME166 SCSI\n");
#endif
	
	if (hostno != -1)
	{
		printk ("ERROR : mvme166_scsi_detect() called twice.\n");
		return 0;
	}

	ncr710 = (NCR710 *)NCR710_BASE;

	/* Initialise ncr710 for low level mode */

	/* Reset chip */

	ncr710->istat = NCR_ISTAT_RST;
	clock=jiffies+2;
	while (jiffies < clock)
		;
	ncr710->istat = 0;

	/* assert  RESET signal on SCSI bus.  */

	ncr710->scntl1 = NCR_SCNTL1_RST;
	clock=jiffies+2;
	while (jiffies < clock)
		;
	ncr710->scntl1 = 0;

	ncr710->scntl0 = NCR_SCNTL0_EPG;

	/* Set low level mode */

	ncr710->dcntl = NCR_DCNTL_CF | NCR_DCNTL_EA | NCR_DCNTL_LLM |
					NCR_DCNTL_COM;

 
	tpnt->this_id = 7;

	instance = scsi_register(tpnt, 0);
	hostno = instance->host_no;
	return 1;
}
	 
const char *mvme166_scsi_info(struct Scsi_Host *shp) {
      static char buffer[256];
        sprintf(buffer, "scsi%d : SCSI at address %p\n", hostno, ncr710);
        return buffer;
}

/*
 * These are our saved pointers for the outstanding command that is 
 * waiting for a reconnect
 */

static unsigned char current_target, current_lun;
static unsigned char *current_cmnd, *current_data;
static int current_nobuffs;
static struct scatterlist *current_buffer;
static int current_bufflen;

static void (*done_fn)(Scsi_Cmnd *) = NULL;
static Scsi_Cmnd * SCint = NULL;


/* 
 * The mvme166_scsi_queue_command() function provides a queued interface
 * to the SCSI driver.  Basically, it just passes control onto the
 * mvme166_scsi_command() function, after fixing it so that the done_fn()
 * is set to the one passed to the function.  We have to be very careful,
 * because there are some commands on some devices that do not disconnect,
 * and if we simply call the done_fn when the command is done then another
 * command is started and queue_command is called again...  We end up
 * overflowing the kernel stack, and this tends not to be such a good idea.
 */

static int recursion_depth = 0;

int mvme166_scsi_queue_command (Scsi_Cmnd * SCpnt,  void (*done)(Scsi_Cmnd *))
	{
	int result;
	Scsi_Cmnd * SCtmp;

	done_fn = done;
	current_target = SCpnt->target;
	current_lun = SCpnt->lun;
	(const void *) current_cmnd = SCpnt->cmnd;
	current_data = (unsigned char *) SCpnt->request_buffer;
	current_bufflen = SCpnt->request_bufflen;
	SCint = SCpnt;
	if(recursion_depth) {
	  return 0;
	};
	recursion_depth++;
	do{
	  result = internal_command (SCint->target, SCint->lun, SCint->cmnd, SCint->request_buffer,
				     SCint->request_bufflen);
	  if (msg_byte(result) == DISCONNECT)  break;
	  SCtmp = SCint;
	  SCint = NULL;
	  SCtmp->result = result;
	  done_fn (SCtmp);
	} while(SCint);
	recursion_depth--;
	return 0;
      }

int mvme166_scsi_command (Scsi_Cmnd * SCpnt) {
	return internal_command (SCpnt->target, SCpnt->lun, SCpnt->cmnd, SCpnt->request_buffer,
				 SCpnt->request_bufflen);
}
	
static int internal_command(unsigned char target, unsigned char lun,
			const void *cmnd, void *buff, int bufflen)
{
	int len = 0;
	unsigned char *data = NULL;
	struct scatterlist *buffer = NULL;
	int nobuffs = 0;
	int clock;			
	int temp;
#if (DEBUG & PHASE_EXIT)
	unsigned char *data_start = NULL;
#endif


#if (DEBUG & PHASE_DATAIN) || (DEBUG & PHASE_DATOUT) 
        int transfered = 0;
#endif

#if (((DEBUG & PHASE_ETC) == PHASE_ETC) || (DEBUG & PRINT_COMMAND) || \
        (DEBUG & PHASE_EXIT))   
        int i;
#endif

#if ((DEBUG & PHASE_ETC) == PHASE_ETC)
        int phase=0, newphase;
#endif

	int done = 0;
	unsigned char status = 0;	
	unsigned char message = 0;
	register unsigned char status_read;

	unsigned transfersize = 0, underflow = 0;

	incommand = 0;
	ncr_aborted = 0;

#if (DEBUG & PRINT_COMMAND)
	printk ("scsi%d : target = %d, command = ", hostno, target);
        print_command((unsigned char *) cmnd);
	printk("\n");
#endif

	if (target == 7)
		return DID_BAD_TARGET;

#if (DEBUG & PHASE_BUS_FREE)
	printk ("scsi%d : phase = BUS FREE \n", hostno);
#endif

/*
 *	BUS FREE PHASE
 *
 * 	On entry, we make sure that the BUS is in a BUS FREE
 *	phase, by insuring that both BSY and SEL are low for
 *	at least one bus settle delay.  Several reads help
 *	eliminate wire glitch.
 */

	clock = jiffies + MVME166_SCSI_BUS_FREE_DELAY;	

	while ((((ncr710->sbcl & STAT_BSY) |
		 (ncr710->sbcl & STAT_BSY) |
		 (ncr710->sbcl & STAT_BSY)) & STAT_BSY) &&
			 (!ncr_aborted) && (jiffies < clock))
		;

	if (jiffies > clock)
		return retcode(DID_BUS_BUSY);
	else if (ncr_aborted)
		return retcode(ncr_aborted);

/*
 *	Bus free has been detected, within BUS settle.
 */

/*
 *	SELECTION PHASE
 *
 *	Now, we select the disk, giving it the SCSI ID as data
 *	and we raise SEL.
 */

#if (DEBUG & PHASE_SELECTION)
	printk("scsi%d : phase = SELECTION\n", hostno);
#endif

	clock = jiffies + MVME166_SCSI_SELECTION_DELAY;

/*
 *	We must assert both our ID and our target's ID on the bus.
 */
	ncr710->sodl = (BYTE) ((1 << target) | 0x80);

	ncr710->scntl1 |= NCR_SCNTL1_ADB;

	ncr710->socl = CMD_SEL;

/*
 *	When the SCSI device decides that we're gawking at it, it will 
 *	respond by asserting BUSY on the bus.
 */
#ifdef DEBUG
	temp = 0;	/* Avoid compiler warning */
#endif

	while (!((status_read = ncr710->sbcl) & STAT_BSY) && 
			(jiffies < clock) && !ncr_aborted)

#if (DEBUG & PHASE_SELECTION)
	{
		temp = clock - jiffies;

		/* What is this????? RGH */

		if (!(jiffies % 5))
			printk("mvme166_scsi_timeout : %d            \r",temp);
	
	}
	printk("Done.                                             \n");
	printk("scsi%d : status = %02x, mvme166_scsi_timeout = %d, aborted = %02x \n", 
			hostno, status_read, temp, ncr_aborted);
#else
		;
#endif
	
	ncr710->scntl1 &= ~NCR_SCNTL1_ADB;

	if ((jiffies > clock)  || (!ncr_aborted && 
			!(status_read & STAT_BSY)))
	{
#if (DEBUG & PHASE_SELECT)
		printk ("scsi%d : NO CONNECT with target %d, status = %x \n", 
				hostno, target, ncr710->sbcl);
#endif
		ncr710->socl = 0;		/* Drop select */
		return retcode(DID_NO_CONNECT);
	}

/*
 *	If we have been aborted, and we have a command in progress, IE the 
 *	target still has BSY asserted, then we will reset the bus, and 
 * 	notify the midlevel driver to expect sense.
 */

	if (ncr_aborted) {
		ncr710->socl = 0;
		if (ncr710->sbcl & STAT_BSY) {
			mvme166_scsi_reset(NULL);
			return retcode(DID_RESET);
		}
		return retcode(ncr_aborted);
	}	

/* Establish current pointers.  Take into account scatter / gather */

        if ((nobuffs = SCint->use_sg)) {
#if (DEBUG & DEBUG_SG)
	    {
		int i;
		printk("scsi%d : scatter gather requested, using %d buffers.\n",
						hostno, nobuffs);
       		buffer = (struct scatterlist *) SCint->buffer;
		for (i = 0; i < nobuffs; ++i, ++buffer)
			printk("scsi%d : buffer %d address = %08x length = %d\n",
				hostno, i, (int)buffer->address, buffer->length);
	    }
#endif

        	buffer = (struct scatterlist *) SCint->buffer;
        	len = buffer->length;
        	data = (unsigned char *) buffer->address;
        }
	else {
#if (DEBUG & DEBUG_SG)
		printk("scsi%d : scatter gather not requested.\n", hostno);
#endif
                buffer = NULL;
                len = SCint->request_bufflen;
                data = (unsigned char *) SCint->request_buffer;
        }
#if (DEBUG & PHASE_EXIT)
	data_start = data;
#endif

#if (DEBUG & (PHASE_DATAIN | PHASE_DATAOUT))
        printk("scsi%d : len = %d\n", hostno, len);
#endif

	/* We are conected.  Set CONNECTED in SCNTL1 */

	ncr710->scntl1 |= NCR_SCNTL1_CON;

	ncr710->socl = 0;		/* Drop select */
	
/*
 * 	INFORMATION TRANSFER PHASE
 *
 *	The nasty looking read / write inline assembler loops we use for 
 *	DATAIN and DATAOUT phases are approximately 4-5 times as fast as 
 *	the 'C' versions - since we're moving 1024 bytes of data, this
 *	really adds up.
 */

#if (DEBUG & PHASE_ETC)
	printk("scsi%d : phase = INFORMATION TRANSFER\n", hostno);
#endif  

	incommand = 1;
	transfersize = SCint->transfersize;
	underflow = SCint->underflow;


/*
 * 	Now, we poll the device for status information,
 *	and handle any requests it makes.  Note that since we are unsure of 
 *	how much data will be flowing across the system, etc and cannot 
 *	make reasonable timeouts, that we will instead have the midlevel
 * 	driver handle any timeouts that occur in this phase.
 */

	while (((status_read = ncr710->sbcl) & STAT_BSY) && !ncr_aborted && !done) 
	{
		if (status_read & STAT_REQ)
		{
#if ((DEBUG & PHASE_ETC) == PHASE_ETC)
			if ((newphase = (status_read & REQ_MASK)) != phase)
			{
				phase = newphase;
				switch (phase)
				{
				case REQ_DATAOUT: 
					printk("scsi%d : phase = DATA OUT\n",
						hostno); 
					break;
				case REQ_DATAIN : 
					printk("scsi%d : phase = DATA IN\n",
						hostno); 
					break;
				case REQ_CMDOUT : 
					printk("scsi%d : phase = COMMAND OUT\n",
						hostno); 
					break;
				case REQ_STATIN :
					 printk("scsi%d : phase = STATUS IN\n",
						hostno); 
					break;
				case REQ_MSGOUT :
					printk("scsi%d : phase = MESSAGE OUT\n",
						hostno); 
					break;
				case REQ_MSGIN :
					printk("scsi%d : phase = MESSAGE IN\n",
						hostno);
					break;
				default : 
					printk("scsi%d : phase = UNKNOWN\n",
						hostno); 
					ncr_aborted = DID_ERROR; 
				}	
			}
#endif
		switch (status_read & REQ_MASK)
		{			
		case REQ_DATAOUT : 
/*
 * 	We loop as long as we are in a data out phase, there is data to send, 
 *	and BSY is still active.
 */
			ncr710->scntl1 |= NCR_SCNTL1_ADB;

			while (len) {
				if (((status_read = ncr710->sbcl) & (STAT_BSY + REQ_MASK)) != (STAT_BSY + REQ_DATAOUT))
					break;
				if (!(status_read & STAT_REQ))
					continue;
				ncr710->sodl = *data++;
				ncr710->socl |= CMD_ACK;
				while (ncr710->sbcl & STAT_REQ)
					;
				ncr710->socl &= ~CMD_ACK;
				len--;
			}
			ncr710->scntl1 &= ~NCR_SCNTL1_ADB;

                        if (!len && nobuffs) {
                                --nobuffs;
                                ++buffer;
                                len = buffer->length;
                                data = (unsigned char *) buffer->address;
#if (DEBUG & DEBUG_SG)
	printk("scsi%d : next scatter-gather buffer len = %d address = %08x\n",
		hostno, len, (int)data);
#endif
                        }
			break;

		case REQ_DATAIN : 
/*
 * 	We loop as long as we are in a data in phase, there is room to read, 
 * 	and BSY is still active
 */
 
			while (len) {
				if (((status_read = ncr710->sbcl) & (STAT_BSY + REQ_MASK)) != (STAT_BSY + REQ_DATAIN))
					break;
				if (!(status_read & STAT_REQ))
					continue;
				*data++ = ncr710->sbdl;
				ncr710->socl |= CMD_ACK;
				while (ncr710->sbcl & STAT_REQ)
					;
				ncr710->socl &= ~CMD_ACK;
				len--;
			}

                        if (!len && nobuffs) {
                                --nobuffs;
                                ++buffer;
                                len = buffer->length;
                                data = (unsigned char *) buffer->address;
#if (DEBUG & DEBUG_SG)
	printk("scsi%d : next scatter-gather buffer len = %d address = %08x\n",
		hostno, len, (int)data);
#endif
                        }

			break;

		case REQ_CMDOUT: 

			ncr710->scntl1 |= NCR_SCNTL1_ADB;

			while (((status_read = ncr710->sbcl) & STAT_BSY) && 
			       ((status_read & REQ_MASK) == REQ_CMDOUT))
				if (status_read & STAT_REQ) {
					ncr710->sodl = *(unsigned char *)cmnd;
					ncr710->socl |= CMD_ACK;
					while (ncr710->sbcl & STAT_REQ)
						;
					ncr710->socl &= ~CMD_ACK;
					cmnd = 1+(unsigned char *) cmnd;
				}

			ncr710->scntl1 &= ~NCR_SCNTL1_ADB;

			break;
	
		case REQ_STATIN : 
			status = ncr710->sbdl;
			ncr710->socl |= CMD_ACK;
			while (ncr710->sbcl & STAT_REQ)
				;
			ncr710->socl &= ~CMD_ACK;
			break;
				
		case REQ_MSGOUT : 
/*
 *	We can only have sent a MSG OUT if we requested to do this 
 *	by raising ATTN.  So, we must drop ATTN.
 */

			ncr710->socl &= ~CMD_ATN;
			ncr710->sodl = NOP;
			ncr710->scntl1 |= NCR_SCNTL1_ADB;
			ncr710->socl |= CMD_ACK;
			while (ncr710->sbcl & STAT_REQ)
				;
			ncr710->socl &= ~CMD_ACK;
			ncr710->scntl1 &= ~NCR_SCNTL1_ADB;
#if (DEBUG & PHASE_MSGOUT)
			printk("scsi%d : sent NOP message.\n", hostno);
#endif
			break;
					
		case REQ_MSGIN : 
			switch (message = ncr710->sbdl) {
			case DISCONNECT :
				done=1;
#if (DEBUG & (PHASE_RESELECT | PHASE_MSGIN))
				printk("scsi%d : disconnected.\n", hostno);
#endif
				break;

			case COMMAND_COMPLETE :
/*
 * Note : we should check for underflow here.   
 */
#if (DEBUG & PHASE_MSGIN)	
				printk("scsi%d : command complete.\n", hostno);
#endif
				done = 1;
				break;
			case ABORT :
#if (DEBUG & PHASE_MSGIN)
				printk("scsi%d : abort message.\n", hostno);
#endif
				done=1;
				break;
			case SAVE_POINTERS :
				current_buffer = buffer;
                                current_bufflen = len;
				current_data = data;
				current_nobuffs = nobuffs;
#if (DEBUG & PHASE_MSGIN)
				printk("scsi%d : pointers saved.\n", hostno);
#endif 
				break;
			case RESTORE_POINTERS:
				buffer=current_buffer;
				cmnd=current_cmnd;
				data=current_data;
				len=current_bufflen;
				nobuffs=current_nobuffs;
#if (DEBUG & PHASE_MSGIN)
				printk("scsi%d : pointers restored.\n", hostno);
#endif
				break;
			default:

/*
 * 	IDENTIFY distinguishes itself from the other messages by setting the
 *	high bit.
 * 	
 *	Note : we need to handle at least one outstanding command per LUN,
 *	and need to hash the SCSI command for that I_T_L nexus based on the 
 *	known ID (at this point) and LUN.
 */

				if (message & 0x80) {
#if (DEBUG & PHASE_MSGIN)
					printk("scsi%d : IDENTIFY message received from id %d, lun %d.\n",
						hostno, target, message & 7);
#endif
				} else {

/*
 *      We should go into a MESSAGE OUT phase, and send  a MESSAGE_REJECT 
 * 	if we run into a message that we don't like.  The driver 
 * 	needs some serious restructuring first though.
 */

#if (DEBUG & PHASE_MSGIN)
					printk("scsi%d : unknown message %d from target %d.\n",
						hostno,  message,   target);
#endif	
				}
			}
			ncr710->socl |= CMD_ACK;
			while (ncr710->sbcl & STAT_REQ)
				;
			ncr710->socl &= ~CMD_ACK;
			break;

		default : 
			printk("scsi%d : unknown phase.\n", hostno); 
			ncr_aborted = DID_ERROR; 
		}	
		} /* if ends */
	} /* while ends */

#if (DEBUG & (PHASE_DATAIN | PHASE_DATAOUT | PHASE_EXIT))
	printk("Transfered %d bytes, allowed %d additional bytes\n", (bufflen - len), len);
#endif

#if (DEBUG & PHASE_EXIT)
	printk("Buffer : \n");
	for (i = 0; i < 20; ++i) 
		printk ("%02x  ", data_start[i]);
	printk("\n");
        printk("scsi%d : status = ", hostno);
        print_status(status);
        printk("message = %02x\n", message);
#endif

	if (ncr_aborted) {
		if (ncr710->sbcl & STAT_BSY) {	
			mvme166_scsi_reset(NULL);
			ncr_aborted = DID_RESET;
		} 
		abort_confirm = 1;
	} 
	ncr710->socl = 0;
	ncr710->scntl1 &= ~(NCR_SCNTL1_CON | NCR_SCNTL1_ADB);

	return retcode (ncr_aborted);
}

int mvme166_scsi_abort (Scsi_Cmnd * SCpnt)
	{
	ncr_aborted = DID_ABORT;

	return SCSI_ABORT_PENDING;
}

/*
	the mvme166_scsi_reset function resets the SCSI bus
*/
	
int mvme166_scsi_reset (Scsi_Cmnd * SCpnt)
{
	unsigned clock;
	/*
		No timeouts - this command is going to fail because 
		it was reset.
	*/

#ifdef DEBUG
	printk("In mvme166_scsi_reset()\n");
#endif


	/* assert  RESET signal on SCSI bus.  */

	ncr710->scntl1 |= NCR_SCNTL1_RST;
	clock=jiffies+2;
	
	/* Wait.  */
	
	while (jiffies < clock)
		;

	ncr710->scntl1 &= ~NCR_SCNTL1_RST;

	ncr_aborted = DID_RESET;

#ifdef DEBUG
	printk("SCSI bus reset.\n");
#endif
        return SCSI_RESET_WAKEUP;
}

#include <asm/segment.h>
#include "sd.h"
#include "scsi_ioctl.h"

int mvme166_scsi_biosparam(Disk *disk, int dev, int* ip) {
  unsigned char buf[256 + sizeof(int) * 2], cmd[6], *data, *page;
  int *sizes, result, formatted_sectors, total_sectors;
  int cylinders, heads, sectors;

/*
 * Only SCSI-I CCS drives and later implement the necessary mode sense 
 * pages.  
 */

  if (disk->device->scsi_level < 2) 
	return -1;

  sizes = (int *) buf;
  data = (unsigned char *) (sizes + 2);

  cmd[0] = MODE_SENSE;
  cmd[1] = (disk->device->lun << 5) & 0xe5;
  cmd[2] = 0x04; /* Read page 4, rigid disk geometry page current values */
  cmd[3] = 0;
  cmd[4] = 255;
  cmd[5] = 0;

/*
 * We are transfering 0 bytes in the out direction, and expect to get back
 * 24 bytes for each mode page.
 */

  sizes[0] = 0;
  sizes[1] = 256;

  memcpy (data, cmd, 6);

  if (!(result = kernel_scsi_ioctl (disk->device, SCSI_IOCTL_SEND_COMMAND, (void *) buf))) {
/*
 * The mode page lies beyond the MODE SENSE header, with length 4, and 
 * the BLOCK DESCRIPTOR, with length header[3].
 */

    page = data + 4 + data[3];
    heads = (int) page[5];
    cylinders = (page[2] << 16) | (page[3] << 8) | page[4];

    cmd[2] = 0x03; /* Read page 3, format page current values */
    memcpy (data, cmd, 6);

    if (!(result = kernel_scsi_ioctl (disk->device, SCSI_IOCTL_SEND_COMMAND, (void *) buf))) {
      page = data + 4 + data[3];
      sectors = (page[10] << 8) | page[11];	

	
/*
 * Get the total number of formatted sectors from the block descriptor, 
 * so we can tell how many are being used for alternates.  
 */

      formatted_sectors = (data[4 + 1] << 16) | (data[4 + 2] << 8) |
	data[4 + 3] ;

      total_sectors = (heads * cylinders * sectors);

/*
 * Adjust the real geometry by subtracting 
 * (spare sectors / (heads * tracks)) cylinders from the number of cylinders.
 *
 * It appears that the CE cylinder CAN be a partial cylinder.
 */

     
printk("scsi%d : heads = %d cylinders = %d sectors = %d total = %d formatted = %d\n",
    hostno, heads, cylinders, sectors, total_sectors, formatted_sectors);

      if (!heads || !sectors || !cylinders)
	result = -1;
      else
        cylinders -= ((total_sectors - formatted_sectors) / (heads * sectors));

/*
 * Now, we need to do a sanity check on the geometry to see if it is 
 * BIOS compatable.  The maximum BIOS geometry is 1024 cylinders * 
 * 256 heads * 64 sectors. 
 */

      if ((cylinders > 1024) || (sectors > 64)) 
	result = -1;
      else {
	ip[0] = heads;
	ip[1] = sectors;
	ip[2] = cylinders;
      }

/* 
 * There should be an alternate mapping for things the mvme166_scsi doesn't
 * understand, but I couldn't say what it is with reasonable certainty.
 */

      }
    }
    
  return result;
}
